import icemac.addressbook.browser.base
import icemac.addressbook.browser.interfaces
import icemac.addressbook.interfaces
import zope.interface


@zope.interface.implementer(
    icemac.addressbook.browser.interfaces.IAddressBookBackground)
class Base(icemac.addressbook.browser.base.BaseView):
    """Base for search result handlers."""

    @property
    def persons(self):
        addressbook = icemac.addressbook.interfaces.IAddressBook(None)
        return [addressbook[id] for id in self.session.get('person_ids', ())]



import pytest
from icemac.addressbook.testing import assert_forbidden


def test_inspector__Inspector__show_context__1(address_book, browser):
    """It renders context information on a view."""
    browser.login('globalmgr')
    browser.open(browser.INSPECTOR_VIEW_URL)
    assert "PersonList u'person-list.html'" in browser.contents
    # the base classes of the view
    assert (
        "(&lt;class 'icemac.addressbook.browser.person.list.PersonList'&gt;, "
        "&lt;class 'z3c.pagelet.browser.BrowserPagelet'&gt;)" in
        browser.contents)
    # an interface of the view
    assert 'z3c.pagelet.interfaces.IPagelet' in browser.contents


def test_inspector__Inspector__show_context__2(address_book, browser):
    """It renders context information on an object."""
    browser.login('globalmgr')
    browser.open(browser.INSPECTOR_OBJECT_URL)
    # Class of context
    assert 'icemac.addressbook.keyword.KeywordContainer' in browser.contents
    # base class of context
    assert 'zope.container.btree.BTreeContainer' in browser.contents
    # an interface of context
    assert 'icemac.addressbook.interfaces.IKeywords' in browser.contents


def test_inspector__Inspector__show_context__3(address_book, browser):
    """It renders context information on a the root object."""
    browser.login('globalmgr')
    browser.open(browser.INSPECTOR_ROOT_OBJECT_URL)
    # the base classes of the root object
    assert (
        "(&lt;class 'zope.container.folder.Folder'&gt;, "
        "&lt;class 'zope.site.site.SiteManagerContainer'&gt;)" in
        browser.contents)
    # an interface of the root object
    assert 'zope.container.interfaces.IContainer' in browser.contents


@pytest.mark.parametrize('username', ('editor', 'visitor', 'mgr'))
def test_inspector__Inspector__1(address_book, browser, username):
    """It cannot be accessed by non-global-admin users on a view."""
    assert_forbidden(browser, username, browser.INSPECTOR_VIEW_URL)


@pytest.mark.parametrize('username', ('editor', 'visitor', 'mgr'))
def test_inspector__Inspector__2(address_book, browser, username):
    """It cannot be accessed by non-global-admin users on an object."""
    assert_forbidden(browser, username, browser.INSPECTOR_OBJECT_URL)


@pytest.mark.parametrize('username', ('editor', 'visitor', 'mgr'))
def test_inspector__Inspector__3(address_book, browser, username):
    """It cannot be accessed by non-global-admin users on the root object."""
    assert_forbidden(browser, username, browser.INSPECTOR_ROOT_OBJECT_URL)



from icemac.addressbook.i18n import _
import datetime
import icemac.addressbook.browser.base
import pytz
import z3c.authviewlet.session
import zope.pluggableauth.plugins.session


class FlashedSessionCredentialsPlugin(
        zope.pluggableauth.plugins.session.SessionCredentialsPlugin,
        icemac.addressbook.browser.base.FlashView):
    """SessionCredentialsPlugin with support for flash messages."""

    def challenge(self, request):
        result = super(FlashedSessionCredentialsPlugin,
                       self).challenge(request)
        self.send_flash(
            _('To log-in enter your username and password and submit the '
              'form.'))
        return result


class FlashedSessionCredentialsLoginForm(
        z3c.authviewlet.session.SessionCredentialsLoginForm,
        icemac.addressbook.browser.base.FlashView):
    """SessionCredentialsLoginForm with flash messages."""

    def update(self):
        super(FlashedSessionCredentialsLoginForm, self).update()
        if str(self.request.response.getStatus()).startswith('3'):
            self.send_flash(_('You have been logged-in successfully.'))
            principals = zope.component.getUtility(
                zope.pluggableauth.interfaces.IAuthenticatorPlugin,
                name=u'icemac.addressbook.principals')
            principal = principals.get(self.request.principal.id)
            principal.last_login = pytz.utc.localize(
                datetime.datetime.utcnow())
        elif 'SUBMIT' in self.request:
            self.send_flash(
                _('Login failed. Username and/or password might be wrong.'))



import icemac.addressbook.browser.interfaces
import z3c.flashmessage.receiver
import zope.component
import zope.contentprovider.provider
import zope.interface


class MessagesContentProvider(
        zope.contentprovider.provider.ContentProviderBase,
        z3c.flashmessage.receiver.GlobalMessageReceiver):
    """Content provider displaying flash messages."""

    zope.component.adapts(
        zope.interface.Interface,
        icemac.addressbook.browser.interfaces.IAddressBookLayer,
        zope.interface.Interface)

    template = zope.browserpage.viewpagetemplatefile.ViewPageTemplateFile(
        'messages.pt')

    def update(self):
        self.messages = list(self.receive())

    def render(self):
        if self.messages:
            return self.template(self)
        return ''



from icemac.addressbook.i18n import _
import icemac.addressbook.interfaces
import zope.schema.fieldproperty
import icemac.addressbook.utils
import persistent
import zope.catalog.interfaces
import zope.component
import zope.container.btree
import zope.container.contained
import zope.interface
import zope.lifecycleevent


class KeywordContainer(zope.container.btree.BTreeContainer):
    "A container for keywords."
    zope.interface.implements(icemac.addressbook.interfaces.IKeywords)

    def get_keywords(self):
        # sorting will be done in presentation layer
        return self.values()

    def get_keyword_by_title(self, title, default=None):
        for keyword in self.values():
            if keyword.title == title:
                return keyword
        return default


class Keyword(persistent.Persistent, zope.container.contained.Contained):
    "A keyword."
    zope.interface.implements(icemac.addressbook.interfaces.IKeyword)
    zope.schema.fieldproperty.createFieldProperties(
        icemac.addressbook.interfaces.IKeyword)

    def __init__(self, title=None):
        super(Keyword, self).__init__()
        if title is not None:
            self.title = title


keyword_entity = icemac.addressbook.entities.create_entity(
    _(u'keyword'), icemac.addressbook.interfaces.IKeyword, Keyword)


@zope.component.adapter(icemac.addressbook.interfaces.IKeyword,
                        zope.lifecycleevent.IObjectModifiedEvent)
def changed(obj, event):
    for desc in event.descriptions:
        if (desc.interface == icemac.addressbook.interfaces.IKeyword and
                'title' in desc.attributes):
            catalog = zope.component.getUtility(
                zope.catalog.interfaces.ICatalog)
            catalog.updateIndex(catalog.get('keywords'))
            break


uniqueTitles = icemac.addressbook.utils.unique_by_attr_factory(
    'title', _(u'This keyword already exists.'))



import BTrees.Length
import icemac.addressbook.namechooser.interfaces
import persistent
import zope.component
import zope.container.contained
import zope.interface


class NameSuffix(persistent.Persistent, zope.container.contained.Contained):
    """Storage for name suffix."""

    zope.interface.implements(
        icemac.addressbook.namechooser.interfaces.INameSuffix)
    zope.component.adapts(
        icemac.addressbook.namechooser.interfaces.IDontReuseNames)

    def __init__(self):
        self._suffix = BTrees.Length.Length(0)

    def __iadd__(self, value):
        self._suffix.change(value)
        return self

    def __unicode__(self):
        return unicode(self._suffix.value)


name_suffix = zope.annotation.factory(
    NameSuffix, key='icemac.namechooser.DontReuseNames.NameSuffix')


class DontReuseNames(zope.container.contained.NameChooser):
    """NameChooser assuring that the chosen names are unique forever.

    It assures this by storing the last chosen suffix in an annotation on
    the container object.

    When a containers provides
    ``icemac.addressbook.namechooser.interfaces.IDontReuseNames`` this name
    chooser is used. (It also needs to provide
    ``zope.annotation.interfaces.IAttributeAnnotatable`` as the information
    gets stored in an annotation.)
    """

    zope.component.adapts(
        icemac.addressbook.namechooser.interfaces.IDontReuseNames)

    def chooseName(self, name, obj):
        # remove characters that checkName does not allow
        name = unicode(name.replace('/', '-').lstrip('+@'))

        if not name:
            name = unicode(obj.__class__.__name__)

        prefix, dot, extension = name.rpartition('.')
        if prefix:
            extension = dot + extension
        else:
            # There is no dot in the name, so
            prefix = extension
            extension = u''

        name_suffix = self.name_suffix
        while True:
            name_suffix += 1
            chosen_name = u'{prefix}-{suffix}{extension}'.format(
                prefix=prefix, suffix=name_suffix, extension=extension)
            if not self.name_in_use(chosen_name):
                break

        # Make sure the name is valid.  We may have started with something bad.
        self.checkName(chosen_name, obj)

        return chosen_name

    def name_in_use(self, name):
        return name in self.context

    @property
    def name_suffix(self):
        return icemac.addressbook.namechooser.interfaces.INameSuffix(
            self.context)



import icemac.addressbook.browser.interfaces
import zope.component
import zope.traversing.namespace


class attr(zope.traversing.namespace.attr):
    """Specialized ++attribute++ traverser."""

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def traverse(self, name, ignored):
        target = super(attr, self).traverse(name, ignored)
        # Iterating the subscribers calls them:
        list(zope.component.subscribers(
            (target, self.request),
            icemac.addressbook.browser.interfaces.IAttributeTraversalHook))
        return target



import icemac.addressbook.generations.utils


def evolve(context):
    """Install local principal annotation utility."""
    icemac.addressbook.generations.utils.update_address_book_infrastructure(
        context)



def test_multi_keyword__Search__1(search_data, browser):
    """`Search` for persons who have specified keywords assigned to."""
    browser.login('visitor')
    browser.open(browser.SEARCH_URL)
    browser.getLink('Keyword search').click()
    assert browser.SEARCH_BY_KEYWORD_URL == browser.url
    # An explanation text is displayed:
    assert (
        ['The keywords you want to search for have to be in the right box. '
         'Use the arrow buttons to move them.'] ==
        browser.etree.xpath(
            '//div[@id="content"]/div[@class="row no-print"]/text()'))


def test_multi_keyword__Search__2(address_book, browser):
    """When no entry in the keywords field is selected a message is shown."""
    browser.login('visitor')
    browser.open(browser.SEARCH_BY_KEYWORD_URL)
    browser.getControl('Search').click()
    assert ('No person found.' == browser.etree.xpath(
        '//div[@id="content"]/div[@class="row"]/text()')[-1].strip())


def test_multi_keyword__Search__3(search_data, browser):
    """When no result is found a message is shown."""
    browser.login('visitor')
    browser.keyword_search('work')
    assert ('No person found.' == browser.etree.xpath(
        '//div[@id="content"]/div[@class="row"]/text()')[-1].strip())


def test_multi_keyword__Search__4(search_data, browser):
    """`Search` result is displayed as a table with links to the edit form."""
    browser.login('visitor')
    browser.keyword_search('church')
    assert (['Koch', 'Liebig', 'Velleuer'] ==
            browser.etree.xpath('//table/tbody/tr/td/a/text()'))
    assert browser.getLink('Koch').url.startswith(browser.PERSON_EDIT_URL)
    # The previously selected keyword is still selected:
    assert browser.getControl('church').selected


def test_multi_keyword__Search__5(search_data, browser):
    """`Search` uses `and` as default search term concatenation."""
    browser.login('visitor')
    browser.open(browser.SEARCH_BY_KEYWORD_URL)
    assert ['and'] == browser.getControl(
        'search term concatenation').displayValue
    browser.getControl('keywords').displayValue = ['church', 'family']
    browser.getControl('Search').click()
    assert (['Koch', 'Velleuer'] ==
            browser.etree.xpath('//table/tbody/tr/td/a/text()'))


def test_multi_keyword__Search__6(search_data, browser):
    """`Search` can also make an or-search of the selected keywords."""
    browser.login('visitor')
    browser.open(browser.SEARCH_BY_KEYWORD_URL)
    browser.getControl('or').click()
    browser.getControl('keywords').displayValue = ['friends', 'family']
    browser.getControl('Search').click()
    assert (['Hohmuth', 'Koch', 'Velleuer'] ==
            browser.etree.xpath('//table/tbody/tr/td/a/text()'))



# -*- coding: latin-1 -*-
import icemac.addressbook.interfaces
import zope.generations.utility
import zope.catalog.interfaces
import zope.component


generation = 6


def evolve(context):
    """icemac.addressbook.interfaces.IKeywords.get_titles has been split into
    IKeywordTitles interface.
    """
    root = zope.generations.utility.getRootFolder(context)
    addressbooks = zope.generations.utility.findObjectsProviding(
        root, icemac.addressbook.interfaces.IAddressBook)
    for addressbook in addressbooks:
        catalog = zope.component.queryUtility(zope.catalog.interfaces.ICatalog,
                                              context=addressbook)

        catalog['keywords'].interface = (
            icemac.addressbook.interfaces.IKeywordTitles)



# -*- coding: utf-8 -*-
from icemac.addressbook.i18n import _
from icemac.addressbook.interfaces import IAddressBook
import icemac.addressbook.browser.base
import z3c.authviewlet.auth
import zope.component
import zope.contentprovider.provider
import zope.pluggableauth.interfaces
import zope.viewlet.viewlet


class AddressBookTitle(zope.contentprovider.provider.ContentProviderBase):
    """Content provider for the addressbook title string."""

    default_title = u'icemac.addressbook'

    def render(self):
        address_book = IAddressBook(self.context)
        if not IAddressBook.providedBy(address_book):
            return self.default_title
        return address_book.title or self.default_title


class LoggedInUserViewlet(
        zope.viewlet.viewlet.ViewletBase,
        icemac.addressbook.browser.base.BaseView):
    """Render info about the user currently logged-in.

    The info contains the log-in name and a link to the principal edit form if
    appropriate.
    """

    @property
    def available(self):
        return z3c.authviewlet.auth.authenticated(self.request.principal)

    def render(self):
        ab_principal = None
        principals = zope.component.queryUtility(
            zope.pluggableauth.interfaces.IAuthenticatorPlugin,
            name=u'icemac.addressbook.principals')
        request_principal = self.request.principal
        if principals is not None:
            ab_principal = principals.get(request_principal.id)
        if ab_principal is None:
            result = _('logged in as "${username}" -',
                       mapping=dict(username=request_principal.title))
        else:
            result = _('logged in as <a href="${url}">${username}</a> -',
                       mapping=dict(url=self.url(ab_principal, 'index.html'),
                                    username=ab_principal.login))
        return zope.i18n.translate(result, context=self.request)



from icemac.addressbook.i18n import _
import icemac.addressbook.addressbook
import icemac.addressbook.browser.base
import icemac.addressbook.browser.interfaces
import icemac.addressbook.browser.menus.menu
import icemac.addressbook.browser.metadata
import icemac.addressbook.interfaces
import z3c.form.button
import zope.interface
import zope.size.interfaces


@zope.interface.implementer(
    icemac.addressbook.browser.interfaces.IAddressBookBackground)
class AddForm(icemac.addressbook.browser.base.BaseAddForm):

    label = _(u'Add new address book')
    interface = icemac.addressbook.interfaces.IAddressBook
    class_ = icemac.addressbook.addressbook.AddressBook
    next_url = 'object'

    def create(self, data):
        self.selected_time_zone = data.pop('time_zone')
        return super(AddForm, self).create(data)

    def add(self, obj):
        super(AddForm, self).add(obj)
        with zope.component.hooks.site(obj):
            obj.time_zone = self.selected_time_zone


@zope.interface.implementer(
    icemac.addressbook.browser.interfaces.IAddressBookBackground)
class EditForm(icemac.addressbook.browser.base.GroupEditForm):

    label = _(u'Edit address book data')
    interface = icemac.addressbook.interfaces.IAddressBook
    groups = (icemac.addressbook.browser.metadata.MetadataGroup,)
    next_url = 'object'
    next_view = 'edit-address_book.html'

    @z3c.form.button.buttonAndHandler(_('Apply'), name='apply')
    def handleApply(self, action):
        super(EditForm, self).handleApply(self, action)

    @z3c.form.button.buttonAndHandler(
        _(u'Delete all persons in address book'), name='delete_content',
        condition=icemac.addressbook.browser.base.can_access(
            '@@delete-address_book-content.html'))
    def handleDeleteContent(self, action):
        self.redirect_to_next_url('object', 'delete-address_book-content.html')


@zope.interface.implementer(
    icemac.addressbook.browser.interfaces.IAddressBookBackground)
class DeleteForm(icemac.addressbook.browser.base.BaseDeleteForm):
    """Delete whole address book."""

    label = _(u'Do you really want to delete this whole address book?')
    interface = icemac.addressbook.interfaces.IAddressBook
    field_names = ('title', )
    next_url_after_cancel = 'parent'

    def _do_delete(self):
        # delete users first
        principals = self.context.principals
        for name in list(principals.keys()):
            del principals[name]
        super(DeleteForm, self)._do_delete()


class PersonCount(object):
    """Adapter to count persons in address book."""

    zope.interface.implements(
        icemac.addressbook.browser.interfaces.IPersonCount)
    zope.component.adapts(icemac.addressbook.interfaces.IAddressBook)

    def __init__(self, address_book):
        basic_unit, self.count = zope.size.interfaces.ISized(
            address_book).sizeForSorting()
        self.notes = zope.i18n.translate(
            _(u'The users inside this address book will not get deleted.'),
            context=zope.globalrequest.getRequest())


@zope.interface.implementer(
    icemac.addressbook.browser.interfaces.IAddressBookBackground)
class DeleteContentForm(icemac.addressbook.browser.base.BaseDeleteForm):
    """Delete address book contents (aka persons)."""

    label = _(
        u'Do you really want to delete all persons in this address book?')
    interface = icemac.addressbook.browser.interfaces.IPersonCount
    next_view = 'edit-address_book.html'

    def _handle_action(self):
        icemac.addressbook.browser.base.delete_persons(
            self.context, self.context.keys())
        self.status = _('Address book contents deleted.')
        self.redirect_to_next_url('object', 'person-list.html')


address_book_views = icemac.addressbook.browser.menus.menu.SelectMenuItemOn(
    'edit-address_book.html', 'delete-address_book-content.html',
    'delete-address_book.html')



from zope.testbrowser.browser import HTTPError, LinkNotFoundError
import pytest


@pytest.mark.webdriver
def test_crud__AddForm__1_webdriver(address_book, webdriver):
    """A new address book can be added and edited."""
    ab = webdriver.address_book
    # Only managers are allowed to create address books:
    webdriver.login('globalmgr', ab.ROOT_URL)
    ab.create()
    ab.title = 'test book'
    ab.assert_default_favicon_selected()
    ab.select_favicon(1)
    ab.timezone = 'Europe/Berlin'
    ab.submit('add')
    assert webdriver.message == '"test book" added.'
    assert ab.ADDRESS_BOOK_2_WELCOME_URL == webdriver.path
    # Editing is done in master data section:
    webdriver.open(ab.ADDRESS_BOOK_EDIT_URL.replace('/ab', '/AddressBook'))
    # The add form actually stored the values:
    assert 'test book' == ab.title
    ab.assert_favicon_selected(1)
    assert "Europe/Berlin" == ab.timezone
    # The edit form is able to change the data:
    ab.title = 'ftest book'
    ab.select_favicon(0)
    ab.submit('apply')
    assert 'Data successfully updated.' == webdriver.message
    # The edit form submits to itself and shows the stored data:
    assert ab.title == 'ftest book'
    ab.assert_favicon_selected(0)
    # The selected time zone shows up in user's preferences:
    prefs = webdriver.prefs
    webdriver.open(prefs.PREFS_URL.replace('/ab', '/AddressBook', 1))
    prefs.open_time_zone_element()
    assert "Europe/Berlin" == prefs.timezone


def test_crud__AddForm__2(address_book, browser):
    """A new address book can be added."""
    browser.login('globalmgr')
    browser.open(browser.ROOT_URL)
    browser.getLink('address book').click()
    assert browser.ADDRESS_BOOK_ADD_URL == browser.url
    browser.getControl('title').value = 'test 2'
    browser.select_favicon()
    browser.getControl('Time zone').displayValue = ['Africa/Juba']
    browser.getControl('Add').click()
    assert '"test 2" added.' == browser.message
    assert browser.ADDRESS_BOOK_2_WELCOME_URL == browser.url


def test_crud__EditForm__1(address_book, browser):
    """An address book can be edited."""
    browser.login('mgr')
    browser.open(browser.ADDRESS_BOOK_EDIT_URL)
    browser.select_favicon()
    browser.getControl('Apply').click()
    assert 'Data successfully updated.' == browser.message


def test_crud__EditForm__2(address_book, browser):
    """Editing the addressbook can be canceled."""
    address_book.title = u'ftest-ab'
    browser.login('mgr')
    browser.open(browser.ADDRESS_BOOK_DEFAULT_URL)
    browser.getLink('Master data').click()
    browser.getLink('Address book').click()
    assert browser.ADDRESS_BOOK_EDIT_URL == browser.url
    assert 'ftest-ab' == browser.getControl('title').value
    browser.getControl('title').value = 'fancy book'
    browser.getControl('Cancel').click()
    assert 'No changes were applied.' == browser.message
    assert 'ftest-ab' == browser.getControl('title').value


@pytest.mark.parametrize("loginname", ('editor', 'visitor'))
def test_crud__EditForm__3(address_book, browser, loginname):
    """Some roles are not allowed to edit address book's data."""
    # There is no link to edit the address book's data (title) because
    # the user is not allowed to do so:
    browser.login(loginname)
    browser.open(browser.MASTER_DATA_URL)
    with pytest.raises(LinkNotFoundError):
        browser.getLink('Address book')
    # Even opening the URL is not possible:
    with pytest.raises(HTTPError) as err:
        browser.open(browser.ADDRESS_BOOK_EDIT_URL)
    assert 'HTTP Error 403: Forbidden' == str(err.value)


def test_crud__DeleteContentForm__1(
        address_book, PersonFactory, UserFactory, PostalAddressFactory,
        browser):
    """The Administrator is able to delete all persons in the address book."""
    # Make sure there are some persons (with addresses) and some users
    # in the address book:
    PersonFactory(address_book, last_name=u'Tester')
    PersonFactory(address_book, last_name=u'Tester 2')
    UserFactory(
        address_book, u'Hans', u'User', u'hans@user.de', u'asdf', ['Visitor'])
    UserFactory(
        address_book, u'Kurt', u'Utzr', u'kurt@utzr.ch', u'asdf', ['Editor'])
    t3 = PersonFactory(address_book, last_name=u'Tester 3')
    PostalAddressFactory(t3, city=u'Hettstedt')

    # To delete all persons in the address book, the Adminstrator has to
    # open the `edit address book` form and select the `Delete all
    # persons in address book` button there. An `are you sure` form is
    # displayed:
    browser.login('mgr')
    browser.open(browser.ADDRESS_BOOK_EDIT_URL)
    browser.getControl('Delete all persons in address book').click()
    assert browser.ADDRESS_BOOK_DELETE_PERSONS_URL == browser.url
    assert ('Do you really want to delete all persons in this address book?' in
            browser.contents)
    # The number of persons in the address book is displayed:
    assert 'class="text-widget int-field">5</span>' in browser.contents
    browser.getControl('Yes').click()
    assert 'Address book contents deleted.' == browser.message
    assert browser.PERSONS_LIST_URL == browser.url
    # Users are not deleted because the are referenced as a safety belt so the
    # current user does not delete his credentials.
    assert 'Person-3">User' in browser.contents
    assert 'Person-4">Utzr' in browser.contents


def test_crud__DeleteContentForm__2(address_book, PersonFactory, browser):
    """`DeleteContentForm` allows to cancel delete."""
    PersonFactory(address_book, last_name=u'Tester')
    browser.login('mgr')
    browser.open(browser.ADDRESS_BOOK_DELETE_PERSONS_URL)
    browser.getControl('No, cancel').click()
    assert 'Deletion canceled.' == browser.message
    assert browser.ADDRESS_BOOK_EDIT_URL == browser.url


@pytest.mark.parametrize("loginname", ('editor', 'visitor'))
def test_crud__DeleteContentForm__3(address_book, browser, loginname):
    """Some roles are not allowed to delete all persons in address book."""
    browser.login(loginname)
    with pytest.raises(HTTPError) as err:
        browser.open(browser.ADDRESS_BOOK_DELETE_PERSONS_URL)
    assert 'HTTP Error 403: Forbidden' == str(err.value)


def test_crud__DeleteForm__1(address_book, UserFactory, browser):
    """`DeleteForm` allows Administrator to delete the whole addressbook."""
    # Users are a bit harder to delete, so we add them to show they can get
    # deleted as well:
    UserFactory(address_book, u'B.', u'Ude', u'b@u.de', u'asdf', ['Visitor'])
    UserFactory(address_book, u'B.', u'Uch', u'b@u.ch', u'asdf', ['Editor'])
    browser.login('globalmgr')
    browser.open(browser.ROOT_URL)
    browser.getLink('Delete').click()
    assert browser.ADDRESS_BOOK_DELETE_URL == browser.url
    browser.getControl('Yes').click()
    assert '"test addressbook" deleted.' == browser.message
    assert browser.ROOT_URL_WITHOUT_SLASH == browser.url
    assert 'There are no address books' in browser.contents


def test_crud__DeleteForm__2(address_book, UserFactory, browser):
    """`DeleteForm` can be canceled."""
    browser.login('mgr')
    browser.open(browser.ADDRESS_BOOK_DELETE_URL)
    browser.getControl('No, cancel').click()
    assert 'Deletion canceled.' == browser.message
    assert browser.ROOT_URL_WITHOUT_SLASH == browser.url


@pytest.mark.parametrize('user', ['visitor', 'editor'])
def test_crud__DeleteForm__3(address_book, browser, user):
    """Non-admin users are not able to access `DeleteForm`."""
    browser.login(user)
    with pytest.raises(HTTPError) as err:
        browser.open(browser.ADDRESS_BOOK_DELETE_URL)
    assert 'HTTP Error 403: Forbidden' == str(err.value)


def test_crud__welcome_pt__1(address_book, browser):
    """The greeting displayed after login contains the address book title."""
    address_book.title = u'ftest-ab'
    browser.login('visitor')
    browser.open(browser.ADDRESS_BOOK_DEFAULT_URL)
    assert ('<p>Welcome to ftest-ab. Please select one of the tabs above.</p>'
            in browser.contents)



def test_name__Search__1(address_book, browser):
    """`Search` for persons who have specified keywords assigned to."""
    browser.login('visitor')
    browser.open(browser.SEARCH_URL)
    browser.getLink('Name search').click()
    assert browser.SEARCH_BY_NAME_URL == browser.url
    # An explanation text is displayed:
    assert (
        ['You may use wildcards in this search: Use ? for a single character '
         'or * for multiple characters.'] ==
        browser.etree.xpath(
            '//div[@id="content"]/div[@class="row no-print"]/text()'))


def test_name__Search__2(address_book, browser):
    """`Search` displays an error message when the search field is empty."""
    browser.login('visitor')
    browser.open(browser.SEARCH_BY_NAME_URL)
    browser.getControl('Search').click()
    assert ['Required input is missing.'] == browser.etree.xpath(
        '//ul[@class="errors"]/li/div/text()')


def test_name__Search__3(address_book, browser):
    """`Search` displays a message when no person is found."""
    browser.login('visitor')
    browser.open(browser.SEARCH_BY_NAME_URL)
    browser.getControl('Name').value = 'Unknown'
    browser.getControl('Search').click()
    assert ('No person found.' == browser.etree.xpath(
        '//div[@id="content"]/div[@class="row"]/text()')[-1].strip())


def test_name__Search__4(search_data, browser):
    """`Search` result is displayed as a table with links to the edit form."""
    browser.login('visitor')
    browser.open(browser.SEARCH_BY_NAME_URL)
    browser.getControl('Name').value = 'Lie*'
    browser.getControl('Search').click()
    assert (['Liebig', 'Tester', 'Liese'] ==  # two search results with 3 links
            browser.etree.xpath('//table/tbody/tr/td/a/text()'))
    assert browser.getLink('Liebig').url.startswith(browser.PERSON_EDIT_URL)
    # The previously entered search string is still there:
    assert 'Lie*' == browser.getControl('Name').value



"""Error renderers"""

from icemac.addressbook.i18n import _
import grokcore.component
import icemac.addressbook.browser
import icemac.addressbook.browser.interfaces
import zope.component
import zope.schema


@grokcore.component.adapter(
    None, zope.schema.interfaces.ConstraintNotSatisfied, name="email")
@grokcore.component.implementer(
    icemac.addressbook.browser.interfaces.IErrorMessage)
def email_constraint_not_satisfield(field, exc):
    value = exc.args[0]
    return _(u'${value} is not a valid e-mail address.',
             mapping=dict(value=value))


def render_error(entity, field_name, exc):
    """Render the error text using the error render adapters."""
    # A queryMultiAdapter call with name=None behaves strange: it seems to
    # delete all adapters matching object and interface, so make sure that
    # the name is not None:
    assert field_name is not None

    # There might be errors which cannot be associated with a specific field
    field = ''
    if field_name:
        field = entity.getField(field_name)

    # try named adapter first
    message = zope.component.queryMultiAdapter(
        (field, exc), icemac.addressbook.browser.interfaces.IErrorMessage,
        name=field_name)
    if message is None:
        # fallback to default (unnamed) adapter
        message = zope.component.getMultiAdapter(
            (field, exc), icemac.addressbook.browser.interfaces.IErrorMessage)
    return message



from icemac.addressbook.i18n import _
import z3c.authviewlet.auth
import icemac.addressbook.browser.base


class FlashedHTTPAuthenticationLogout(
        z3c.authviewlet.auth.HTTPAuthenticationLogout,
        icemac.addressbook.browser.base.FlashView):
    """HTTPAuthenticationLogout enriched by flash messages."""

    def logout(self, nextURL=None):
        result = super(FlashedHTTPAuthenticationLogout, self).logout(nextURL)
        self.send_flash(_('You have been logged out successfully.'))
        return result



import icemac.addressbook.generations.utils


@icemac.addressbook.generations.utils.evolve_addressbooks
def evolve(ab):
    """Fix selected startpage after adding by default to view names."""
    startpage_view = ab.startpage[1]
    if startpage_view is not None and startpage_view.startswith('@@'):
        ab.startpage = (ab.startpage[0], ab.startpage[1][2:])



# -*- coding: utf-8 -*-
import icemac.addressbook.interfaces
import zc.sourcefactory.basic
import zc.sourcefactory.contextual
import zope.component
import zope.security.proxy
import zope.securitypolicy.interfaces


class RoleSource(zc.sourcefactory.basic.BasicSourceFactory):
    """A source for all global roles."""

    def getValues(self):
        roles = zope.component.getAllUtilitiesRegisteredFor(
            zope.securitypolicy.interfaces.IRole)
        names = [r.id
                 for r in roles
                 if (r.id.startswith('icemac.addressbook.global.') or
                     r.id.startswith('icemac.ab.'))]
        return sorted(names)

    def getTitle(self, value):
        role = zope.component.getUtility(
            zope.securitypolicy.interfaces.IRole, name=value)
        return role.title


role_source = RoleSource()


class Persons(zc.sourcefactory.contextual.BasicContextualSourceFactory):
    """Persons in addressbook."""

    def getValues(self, context):
        root = icemac.addressbook.principals.interfaces.IRoot(context)
        insecured_context = zope.security.proxy.removeSecurityProxy(context)
        # When editing a user we need this users person in list, as
        # the person on user is not changeable we need only to return
        # this person.
        if isinstance(insecured_context,  # XXX Use IPrincipal.providedBy?
                      icemac.addressbook.principals.principals.Principal):
            values = [context.person]
        else:
            values = []
            persons_of_existing_pricipals = frozenset(
                [principal.person for principal in context.values()])

            for person in root.values():
                if not getattr(person.default_email_address, 'email', None):
                    # Show only persons which have an e-mail address:
                    continue
                if person in persons_of_existing_pricipals:
                    # Show only persons which are not yet users:
                    continue
                values.append(person)
        return values

    def getTitle(self, context, value):
        return icemac.addressbook.interfaces.ITitle(value)


persons = Persons()



from zope.principalannotation.interfaces import IPrincipalAnnotationUtility
import zope.component
import zope.component.hooks
import zope.generations.utility


def evolve(context):
    """Remove global principal annotations utility.

    Data was migrated in the previous generation.

    """
    root_folder = zope.generations.utility.getRootFolder(context)
    try:
        old_site = zope.component.hooks.getSite()
        zope.component.hooks.setSite(root_folder)

        util = zope.component.getUtility(IPrincipalAnnotationUtility)
        sm = root_folder.getSiteManager()
        assert sm.unregisterUtility(
            provided=IPrincipalAnnotationUtility) is True
        del util.__parent__[util.__name__]
    finally:
        zope.component.hooks.setSite(old_site)



# -*- coding: utf-8 -*-
from icemac.addressbook.interfaces import IPerson, IPostalAddress
import decimal
import pytest


KEYWORD = u'keyword for test'


# Fixtures to create objects:

@pytest.fixture(scope='function')
def UpdateablePersonFactory(FullPersonFactory):
    """Return a callable which creates a person object that can be updated.

    Supports user defined fields and sets a default keyword if `keywords` is
    not in the parameters of the call.

    """
    def create_updateable_person(address_book, **kw):
        kw.setdefault('keywords', [KEYWORD])
        kw.setdefault('last_name', u'Tester')
        return FullPersonFactory(address_book, **kw)
    return create_updateable_person


@pytest.fixture(scope='function')
def AddressWithUserdefinedFieldFactory(
        FieldFactory, UpdateablePersonFactory, PostalAddressFactory):
    """Callable to create an address on a person with a user defined field."""
    def _create_user_defined_field(address_book, field_type, field_value):
        """Create a user defined field."""
        field_name = FieldFactory(
            address_book, IPostalAddress, field_type, u'distance').__name__
        return PostalAddressFactory(
            UpdateablePersonFactory(address_book),
            **{field_name: field_value, 'set_as_default': True})
    return _create_user_defined_field


# Helper functions

def _update_field_value(browser, field_name, operator, value):
    """Update a field using the update search result handler."""
    browser.login('mgr')
    browser.keyword_search(KEYWORD, apply='Update')
    browser.getControl('field').displayValue = [field_name]
    browser.getControl('Next').click()
    assert '' == browser.getControl('new value', index=0).value
    browser.getControl('new value', index=0).value = value
    browser.getControl('operation').displayValue = [operator]
    browser.getControl('Next').click()


# Tests

def test_update__endtoend__1(search_data, browser):
    """Testing updating selected persons end to end.

    The `update` search result handler allows to update a single field on each
    selected person.

    """
    # The `searchDataS` fixture defines some persons. When user searches for
    # them all persons are selected by default so he only has to select the
    # `update` search handler to perform a multi-update:
    browser.login('mgr')
    browser.keyword_search('family', apply='Update')

    # The user is guided through the update using a wizard.
    # 1st) Choose a field for update:
    assert ['person -- first name', 'person -- last name',
            'person -- birth date'] == browser.getControl(
                'field').displayOptions[:3]
    browser.getControl('field').displayValue = ['person -- notes']
    browser.getControl('Next').click()

    # 2nd) Enter a new value for the selected field and choose an operation
    # which defaults to 'append':
    assert ['append new value to existing one'] == browser.getControl(
        'operation').displayValue
    browser.getControl('new value', index=0).value = '\tfoobar'
    browser.getControl('Next').click()

    # 3rd) Check result:
    assert 2 == browser.contents.count('\tfoobar')

    # 4th) Hitting `Complete` persists the change and redirects to the person
    # list, displaying a message:
    browser.getControl('Complete').click()
    assert browser.PERSONS_LIST_URL == browser.url
    assert 'Data successfully updated.' == browser.message

    # The fields got changed as promised in the message:
    browser.getLink('Person list').click()
    browser.getLink('Koch').click()
    assert 'father-in-law\tfoobar' == browser.getControl('notes').value


def test_update__endtoend__2(search_data, browser):
    """Adding an empty new value does not change the updated value.

    This is an edge case test.

    """
    browser.login('mgr')
    browser.keyword_search('family', apply='Update')
    browser.getControl('field').displayValue = ['person -- last name']
    browser.getControl('Next').click()
    browser.getControl('new value', index=0).value = ''
    browser.getControl('operation').displayValue = [
        'append new value to existing one']
    browser.getControl('Next').click()
    # The last name column is displayed as a link column it contains the
    # unchanged last name:
    assert ('<td><a href="http://localhost/ab/Person-2">Koch</a></td>' in
            browser.contents)


def test_update__endtoend__3(
        address_book, FieldFactory, UpdateablePersonFactory, browser):
    """A user defined boolean field can be updated."""
    field_name = FieldFactory(
        address_book, IPerson, 'Bool', u'Ever met').__name__
    UpdateablePersonFactory(address_book, **{field_name: False})
    browser.login('mgr')
    browser.keyword_search(KEYWORD, apply='Update')
    browser.getControl('field').displayValue = ['person -- Ever met']
    browser.getControl('Next').click()
    browser.getControl('yes').click()
    browser.getControl('operation').displayValue = [
        'replace existing value with new one']
    browser.getControl('Next').click()
    # Update sets the value to 'yes':
    assert '<td>Tester</td><td>yes</td>' in browser.contents_without_whitespace


def test_update__endtoend__4(
        address_book, FieldFactory, UpdateablePersonFactory,
        PostalAddressFactory, browser):
    """A user defined choice field can be updated."""
    field_name = FieldFactory(
        address_book, IPostalAddress, 'Choice', u'distance',
        values=[u'< 50 km', u'>= 50 km']).__name__
    PostalAddressFactory(UpdateablePersonFactory(address_book),
                         **{field_name: '>= 50 km', 'set_as_default': True})

    browser.login('mgr')
    browser.keyword_search(KEYWORD, apply='Update')
    browser.getControl('field').displayValue = ['postal address -- distance']
    browser.getControl('Next').click()
    assert ['No value', '< 50 km', '>= 50 km'] == browser.getControl(
        'new value').displayOptions
    browser.getControl('new value').displayValue = ['< 50 km']
    browser.getControl('operation').displayValue = [
        'replace existing value with new one']
    browser.getControl('Next').click()
    # Update sets the value to '< 50 km':
    assert ('<td>Tester</td><td><50km</td>' in
            browser.contents_without_whitespace)


def test_update__endtoend__5(address_book, UpdateablePersonFactory, browser):
    """The keywords field can be updated."""
    UpdateablePersonFactory(address_book, keywords=[KEYWORD, u'second kw'])
    browser.login('mgr')
    browser.keyword_search('second kw', apply='Update')
    browser.getControl('field').displayValue = ['person -- keywords']
    browser.getControl('Next').click()
    assert [KEYWORD, 'second kw'] == browser.getControl(
        'new value').displayOptions
    browser.getControl('new value').displayValue = ['second kw']
    browser.getControl('operation').displayValue = [
        'remove selected keywords from existing ones']
    browser.getControl('Next').click()
    assert ('<td>Tester</td><td>keywordfortest</td>' in
            browser.contents_without_whitespace)
    browser.getControl('Complete').click()
    # After removing the keyword from the person no person can be found:
    browser.keyword_search(u'second kw')
    assert 'No person found.' in browser.contents


@pytest.mark.parametrize("type,factory", (
    ['Int', int],
    ['Decimal', decimal.Decimal]))
def test_update__endtoend__6(
        address_book, AddressWithUserdefinedFieldFactory, browser, type,
        factory):
    """A user defined integer field can be updated."""
    AddressWithUserdefinedFieldFactory(address_book, type, factory(50))
    _update_field_value(browser, 'postal address -- distance', 'add', '5')
    assert '<td>Tester</td><td>55</td>' in browser.contents_without_whitespace


def test_update__endtoend__7(
        address_book, UpdateablePersonFactory, EMailAddressFactory, browser):
    """Validation errors show up in the result table."""
    EMailAddressFactory(
        UpdateablePersonFactory(address_book), set_as_default=True)
    _update_field_value(
        browser, 'e-mail address -- e-mail address', 'append', 'foo')
    assert ('<td>Tester</td><td></td><td>fooisnotavalide-mailaddress.</td>' in
            browser.contents_without_whitespace)
    # Complete button is not shown:
    assert ['form.buttons.back'] == browser.submit_control_names


def test_update__endtoend__8(
        address_book, AddressWithUserdefinedFieldFactory, browser):
    """Division by zero is handled like a validation error."""
    AddressWithUserdefinedFieldFactory(address_book, 'Int', 50)
    _update_field_value(browser, 'postal address -- distance', 'div', '0')
    assert ('<td>Tester</td><td>50</td><td>Divisionbyzero</td>' in
            browser.contents_without_whitespace)
    # Complete button is not shown:
    assert ['form.buttons.back'] == browser.submit_control_names


def test_update__endtoend__9(address_book, UpdateablePersonFactory, browser):
    """Field selected for change can be changed."""
    UpdateablePersonFactory(address_book)
    _update_field_value(browser, 'person -- first name', 'append', 'foo')
    browser.getControl('Back').click()
    browser.getControl('Back').click()
    browser.getControl('field').displayValue = ['person -- birth date']
    browser.getControl('Next').click()
    # The newly selected field for update might have a data types which does
    # not match the previous selected on, so the value is deleted
    assert '' == browser.getControl('new value').value


def test_update__endtoend__10(address_book, UpdateablePersonFactory, browser):
    """Not hitting the `complete button` does not persist any changes."""
    UpdateablePersonFactory(address_book)
    _update_field_value(browser, 'person -- last name', 'append', 'foo')
    browser.getLink('Person list').click()
    # The last name of person 'Tester' is unchanged:
    assert ('<a href="{0.PERSON_EDIT_URL}">Tester</a>'.format(browser) in
            browser.contents)


def test_update__endtoend__11(address_book, UpdateablePersonFactory, browser):
    """If the user selects a step with data missing he gets redirected back."""
    UpdateablePersonFactory(address_book)
    browser.login('mgr')
    browser.keyword_search(KEYWORD, apply='Update')
    assert browser.SEARCH_MULTI_UPDATE_URL == browser.url
    browser.getLink('New value').click()
    # 'chooseField' is the first step, so we get redirected there
    assert browser.SEARCH_MULTI_UPDATE_CHOOSE_FIELD_URL == browser.url
    browser.getLink('Check result').click()
    assert browser.SEARCH_MULTI_UPDATE_CHOOSE_FIELD_URL == browser.url
    browser.getControl('field').displayValue = ['person -- first name']
    browser.getControl('Next').click()
    assert browser.SEARCH_MULTI_UPDATE_ENTER_VALUE_URL == browser.url
    browser.getLink('Choose field').click()
    assert browser.SEARCH_MULTI_UPDATE_CHOOSE_FIELD_URL == browser.url
    browser.getLink('Check result').click()
    # After selecting the field 'enterValue' is complete, as the only
    # required field has a default value
    assert browser.SEARCH_MULTI_UPDATE_CHECK_RESULT_URL == browser.url
    # There is no 'complete' button as the user did not enter data in step
    # 'enterValue':
    assert ['form.buttons.back'] == browser.submit_control_names



import icemac.addressbook.generations.utils
import icemac.addressbook.interfaces
import zope.component


@icemac.addressbook.generations.utils.evolve_addressbooks
def evolve(ab):
    """Update user generated fields.

    They get the information about the entity interface they belong to.

    """
    entities = zope.component.getUtility(
        icemac.addressbook.interfaces.IEntities)
    for entity in entities.getEntities(sorted=False):
        for name, field in entity.getRawFields(sorted=False):
            if icemac.addressbook.interfaces.IField.providedBy(field):
                field.interface = entity.interface



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from zope.testbrowser.browser import HTTPError, LinkNotFoundError
import datetime
import pytest
import pytz
import zope.component.hooks


@pytest.yield_fixture(scope='function')
def localadmin(address_book, UserFactory, browser):
    """Create an administrator inside the address book and log in him.

    Returns a browser instance.

    """
    # Using an administrator defined inside the address book to show that he
    # serves fine as administrator.
    UserFactory(address_book, 'Arne', 'Admin', 'arne@example.com', '1qay2wsx',
                ['Administrator'])
    browser.formlogin('arne@example.com', '1qay2wsx')
    # Restore site setting which the browser usage has resetted:
    with zope.component.hooks.site(address_book):
        yield browser


def test_principals__Overview__1(address_book, browser):
    """If no users are defined `Overview` shows a message telling that."""
    browser.login('mgr')
    browser.open(browser.PRINCIPALS_LIST_URL)
    # The users list is initially empty because `mgr` belongs to the global
    # users:
    assert ('No users defined yet or you are not allowed to access any' in
            browser.contents)


def test_principals__Overview__2(address_book, UserFactory, browser):
    """The roles displayed on the overview list are listed comma separated."""
    UserFactory(address_book, 'Hans', 'Tester', 'tester@example.com', 'secret',
                ['Visitor', 'Editor'])
    # The created user cannot login via BasicAuth and `editor` is not allowed
    # to see him, so we have to use `mgr`.
    browser.login('mgr')
    browser.open(browser.PRINCIPALS_LIST_URL)
    assert ['Editor, Visitor'] == browser.etree.xpath('//tr/td[3]/text()')


def test_principals__Overview__3(
        localadmin, address_book, UserFactory, browser):
    """`Overview` displays users sorted by the name column."""
    UserFactory(address_book, u'Ben', u'Utzer', u'ben.utzer@example.com',
                u'12345678', ['Visitor'])
    UserFactory(address_book, u'C.', u'Liebig', u'b@example.com',
                u'12345678', ['Visitor'])
    browser = localadmin.open(localadmin.PRINCIPALS_LIST_URL)
    assert ['Admin, Arne',
            'Liebig, C.',
            'Utzer, Ben'] == browser.etree.xpath(
                '//table/tbody/tr/td/a/text()')


@pytest.mark.parametrize('role', [u'Visitor', 'Editor'])
def test_principals__Overview__4(address_book, UserFactory, browser, role):
    """`Overview` shows a non-admin only one user: himself."""
    UserFactory(address_book, u'Urs', u'Unstable', u'uu@example.com',
                'u1u2u3u4', [role])
    UserFactory(address_book, u'Urs2', u'Unstable2', u'uu2@example.com',
                'u1u2u3u4', [role])
    browser.formlogin(u'uu@example.com', 'u1u2u3u4')
    browser.open(browser.PRINCIPALS_LIST_URL)
    assert (['Unstable, Urs'] ==
            browser.etree.xpath('//table/tbody/tr/td/a/text()'))


def test_principals__Overview__5(address_book, UserFactory, browser):
    """`Overview` truncates too long notes contents."""
    user = UserFactory(
        address_book, u'Ben', u'Utzer', u'ben.utzer@example.com', u'12345678',
        ['Visitor'])
    user.description = (u'This is a new user created to show the '
                        u'truncation of too long notes in the `Overview`.')
    browser.login('mgr')
    browser.open(browser.PRINCIPALS_LIST_URL)
    assert ([u'This is a new user created to show the truncation of …'] ==
            browser.etree.xpath('//table/tbody/tr/td[5]/text()'))


def test_principals__Overview__5_5(
        address_book, UserFactory, TimeZonePrefFactory, browser):
    """It renders the last login time in user's local time zone."""
    user = UserFactory(
        address_book, u'Ben', u'Utzer', u'ben.utzer@example.com', u'12345678',
        ['Visitor'])
    user.last_login = pytz.utc.localize(datetime.datetime(2016, 6, 20, 18, 13))
    TimeZonePrefFactory('Europe/Berlin')
    browser.login('mgr')
    browser.open(browser.PRINCIPALS_LIST_URL)
    assert (['16/06/20 20:13'] ==
            browser.etree.xpath('//table/tbody/tr/td[4]/text()'))


@pytest.mark.parametrize('user', ['visitor', 'editor'])
def test_principals__Overview__6(address_book, browser, user):
    """There is no link to create a user in `Overview` for non-admin users."""
    browser.login(user)
    browser.open(browser.PRINCIPALS_LIST_URL)
    with pytest.raises(LinkNotFoundError):
        browser.getLink('user')


def test_principals__AddForm__1(
        localadmin, address_book, FullPersonFactory, browser2):
    """`AddForm` allows Administrators to create new users."""
    FullPersonFactory(address_book, u'Tester', first_name=u'Hans',
                      email__email=u'tester@example.com')
    FullPersonFactory(address_book, u'Utzer', first_name=u'Ben',
                      email__email=u'ben.utzer@example.com')
    FullPersonFactory(address_book, u'Liebig', first_name=u'B.')
    browser = localadmin.open(localadmin.PRINCIPALS_LIST_URL)
    browser.getLink('user').click()
    assert browser.PRINCIPAL_ADD_URL == browser.url
    # Only persons with a default e-mail address who are not yet users are
    # shown as possible users. (So 'B. Liebig' (no mails address) and 'Admin,
    # Arne' (already a user) are not shown.):
    assert (['Tester, Hans', 'Utzer, Ben'] ==
            browser.getControl('person').displayOptions)
    browser.getControl('person').displayValue = ['Tester, Hans']
    browser.getControl('password', index=0).value = '1234567890'
    browser.getControl('password repetition').value = '1234567890'
    browser.getControl('Visitor').click()
    browser.getControl('Add').click()
    assert '"Tester, Hans" added.' == browser.message
    assert browser.PRINCIPALS_LIST_URL == browser.url
    assert (['Admin, Arne', 'Tester, Hans'] ==
            browser.etree.xpath('//td/a/text()'))

    browser2.formlogin('tester@example.com', '1234567890')
    assert 'You have been logged-in successfully.' == browser2.message
    assert browser.ADDRESS_BOOK_WELCOME_URL == browser2.url


def test_principals__AddForm__2(localadmin):
    """`AddForm` can be cancelled."""
    browser = localadmin.open(localadmin.PRINCIPAL_ADD_URL)
    browser.getControl('Cancel').click()
    assert 'Addition canceled.' == browser.message
    assert browser.PRINCIPALS_LIST_URL == browser.url


def test_principals__AddForm__3(localadmin, address_book, FullPersonFactory):
    """`AddForm` validates that password matches its repetition."""
    FullPersonFactory(address_book, u'Tester', first_name=u'Hans',
                      email__email=u'tester@example.com')
    browser = localadmin.open(localadmin.PRINCIPAL_ADD_URL)
    browser.getControl('person').displayValue = ['Tester, Hans']
    browser.getControl('password', index=0).value = '1234567890'
    browser.getControl('password repetition').value = '0123456789'
    browser.getControl('Add').click()
    assert [] == browser.message
    assert (['Entry in password field was not equal to entry in password '
             'repetition field.'] ==
            browser.etree.xpath('//div[@class="error"]/text()'))


def test_principals__AddForm__4(localadmin, address_book, FullPersonFactory):
    """`AddForm` validates that the password is not too short."""
    FullPersonFactory(address_book, u'Tester', first_name=u'Hans',
                      email__email=u'tester@example.com')
    browser = localadmin.open(localadmin.PRINCIPAL_ADD_URL)
    browser.getControl('person').displayValue = ['Tester, Hans']
    browser.getControl('password', index=0).value = '123'
    browser.getControl('password repetition').value = '123'
    browser.getControl('Add').click()
    assert [] == browser.message
    assert (['Value is too short', 'Value is too short'] ==
            browser.etree.xpath(
                '//ul[@class="errors"]/li/div[@class="error"]/text()'))


def test_principals__AddForm__5(localadmin, address_book, FullPersonFactory):
    """`AddForm` makes sure the login name of the principals stays unique."""
    FullPersonFactory(address_book, u'Admin2', first_name=u'Arne2',
                      email__email=u'arne@example.com')
    browser = localadmin.open(localadmin.PRINCIPAL_ADD_URL)
    browser.getControl('person').displayValue = ['Admin2, Arne2']
    browser.getControl('password', index=0).value = '12345678'
    browser.getControl('password repetition').value = '12345678'
    browser.getControl('Add').click()
    assert [] == browser.message
    assert (['Principal Login already taken!'] ==
            browser.etree.xpath('//div[@class="error"]/text()'))


@pytest.mark.parametrize('user', ['visitor', 'editor'])
def test_principals__AddForm__6(address_book, browser, user):
    """`AddForm` cannot be accessed by a visitor or an editor."""
    browser.login(user)
    with pytest.raises(HTTPError) as err:
        browser.open(browser.PRINCIPAL_ADD_URL)
    assert 'HTTP Error 403: Forbidden' == str(err.value)


def test_principals__AddForm__7(address_book, UserFactory, localadmin):
    """`AddForm` can create a new user for a person those user was deleted."""
    UserFactory(address_book, u'Urs', u'Unstable', u'uu@example.com',
                'u1u2u3u4', ['Visitor'])
    browser = localadmin.open(localadmin.PRINCIPAL_DELETE_URL)
    browser.getControl('Yes').click()
    assert '"Unstable, Urs" deleted.' == browser.message
    browser.open(browser.PRINCIPAL_ADD_URL)
    assert ['Unstable, Urs'] == browser.getControl('person').displayOptions
    browser.getControl('person').displayValue = ['Unstable, Urs']
    browser.getControl('password', index=0).value = '09876543'
    browser.getControl('password', index=1).value = '09876543'
    browser.getControl('Add').click()
    assert '"Unstable, Urs" added.' == browser.message


def test_principals__EditForm__1(localadmin, address_book, UserFactory):
    """`EditForm` allows to edit a principal."""
    UserFactory(address_book, 'Hans', 'Tester', 'hans@example.com',
                '1234567890', ['Visitor'])
    browser = localadmin.open(localadmin.PRINCIPALS_LIST_URL)
    browser.getLink('Tester, Hans').click()
    assert browser.PRINCIPAL_EDIT_URL == browser.url
    assert 'hans@example.com' == browser.getControl('login').value
    # The password is not displayed.
    assert '' == browser.getControl('password', index=0).value
    assert '' == browser.getControl('password repetition').value
    assert browser.getControl('Visitor').selected
    browser.getControl('notes').value = 'Hans the tester'
    browser.getControl('login').value = 'hans@example.com'
    browser.getControl('roles').displayValue = ['Editor']
    browser.getControl('Apply').click()
    assert 'Data successfully updated.' == browser.message
    assert browser.PRINCIPALS_LIST_URL == browser.url
    assert 'Hans the tester' in browser.contents
    browser.open(browser.PRINCIPAL_EDIT_URL)
    assert 'Hans the tester' == browser.getControl('notes').value
    assert 'hans@example.com' == browser.getControl('login').value
    assert ['Editor'] == browser.getControl('roles').displayValue


def test_principals__EditForm__2(localadmin, address_book, UserFactory):
    """`EditForm` allows to cancel editing."""
    UserFactory(address_book, 'Hans', 'Tester', 'hans@example.com',
                '1234567890', ['Visitor'])
    browser = localadmin.open(localadmin.PRINCIPAL_EDIT_URL)
    browser.getControl('notes').value = '2nd new visitor'
    browser.getControl('Cancel').click()
    assert 'No changes were applied.' == browser.message
    assert browser.PRINCIPALS_LIST_URL == browser.url
    assert '2nd new visitor' not in browser.contents


def test_principals__EditForm__3(localadmin, address_book, UserFactory):
    """`EditForm` makes sure the login name of the principals stays unique."""
    UserFactory(address_book, 'Hans', 'Tester', 'hans@example.com',
                '1234567890', ['Visitor'])
    browser = localadmin.open(localadmin.PRINCIPAL_EDIT_URL)
    browser.getControl('login').value = 'arne@example.com'
    browser.getControl('Apply').click()
    assert [] == browser.message
    assert (['Principal Login already taken!'] ==
            browser.etree.xpath('//div[@class="error"]/text()'))


def test_principals__EditForm__4(localadmin):
    """Changing the mail address in persons data does change the login name."""
    browser = localadmin.open(localadmin.PERSON_EDIT_URL)
    browser.getControl('e-mail address', index=1).value = 'htester@example.com'
    browser.getControl('Apply').click()
    assert 'Data successfully updated.' == browser.message
    browser.open(browser.PRINCIPAL_EDIT_URL_1)
    assert 'arne@example.com' == browser.getControl('login').value


def test_principals__EditForm__5(address_book, UserFactory, browser):
    """A visitor can only change his password in the `EditForm`."""
    UserFactory(address_book, u'Urs', u'Unstable', u'uu@example.com',
                'u1u2u3u4', ['Visitor'])
    browser.formlogin(u'uu@example.com', 'u1u2u3u4')
    browser.open(browser.PRINCIPAL_EDIT_URL_1)
    # There are only the password fields and submit controls, but there is no
    # delete button:
    assert ['form.buttons.apply',
            'form.buttons.cancel',
            'form.widgets.password',
            'form.widgets.password_repetition'] == browser.all_control_names
    browser.getControl('password', index=0).value = 'new_password'
    browser.getControl('password repetition').value = 'new_password'
    browser.getControl('Apply').click()
    # Changing the password immediately asks the visitor for his new
    # credentials:
    assert (
        ['You changed the password, please re-login.',
         'Data successfully updated.',
         'To log-in enter your username and password and submit the form.'] ==
        browser.message)
    assert browser.url.startswith(browser.LOGIN_BASE_URL)
    browser.formlogin(u'uu@example.com', 'new_password', use_current_url=True)
    # The new password can also be used to "normally" log-in:
    browser.getLink('Logout').click()
    browser.html_redirect()
    assert (
        ['You have been logged out successfully.',
         'To log-in enter your username and password and submit the form.'] ==
        browser.message)
    browser.formlogin('uu@example.com', 'new_password', use_current_url=True)
    assert browser.url.startswith(browser.PRINCIPALS_LIST_URL)  # +/index.html


def test_principals__EditForm__5_5(address_book, UserFactory, browser):
    """If a visitor does not change his password, he is not logged out."""
    UserFactory(address_book, u'Urs', u'Unstable', u'uu@example.com',
                'u1u2u3u4', ['Visitor'])
    browser.formlogin(u'uu@example.com', 'u1u2u3u4')
    browser.open(browser.PRINCIPAL_EDIT_URL_1)
    browser.getControl('Apply').click()
    # Changing the password immediately asks the visitor for his new
    # credentials:
    assert 'Data successfully updated.' == browser.message
    assert browser.PRINCIPALS_LIST_URL == browser.url


def test_principals__EditForm__6(address_book, UserFactory, browser):
    """An editor can edit his own user data but not the roles."""
    user = UserFactory(address_book, u'Urs', u'Unstable', u'uu@example.com',
                       'u1u2u3u4', ['Editor'])
    user.description = u'Hans the tester'
    browser.formlogin(u'uu@example.com', 'u1u2u3u4')
    browser.open(browser.PRINCIPAL_EDIT_URL_1)
    # There are all fields besides the roles field and the submit controls,
    # but there is no delete button:
    assert ['form.buttons.apply',
            'form.buttons.cancel',
            'form.widgets.description',
            'form.widgets.login',
            'form.widgets.password',
            'form.widgets.password_repetition'] == browser.all_control_names
    assert 'uu@example.com' == browser.getControl('login').value
    assert '' == browser.getControl('password', index=0).value
    assert 'Hans the tester' == browser.getControl('notes').value
    # The editor can change the data and save them. The login-name needn't
    # even be an e-mail address. Changing the own log-in name immediately
    # asks the user for his new credentials:
    browser.getControl('login').value = 'hans.tester'
    browser.getControl('password', index=0).value = 'testtest'
    browser.getControl('password repetition').value = 'testtest'
    browser.getControl('notes').value = 'The big HANS'
    browser.getControl('Apply').click()
    assert (
        ['You changed the login name, please re-login.',
         'You changed the password, please re-login.',
         'Data successfully updated.',
         'To log-in enter your username and password and submit the form.'] ==
        browser.message)
    assert browser.url.startswith(browser.LOGIN_BASE_URL)
    browser.formlogin('hans.tester', 'testtest', use_current_url=True)
    assert browser.url.startswith(browser.PRINCIPALS_LIST_URL)  # +/index.html
    # The data was stored:
    browser.getLink('Unstable').click()
    assert 'hans.tester' == browser.getControl('login').value
    assert '' == browser.getControl('password', index=0).value
    assert 'The big HANS' == browser.getControl('notes').value
    # The new log-in name an password can also be used to "normally" log-in:
    browser.getLink('Logout').click()
    browser.html_redirect()
    assert (
        ['You have been logged out successfully.',
         'To log-in enter your username and password and submit the form.'] ==
        browser.message)
    browser.formlogin('hans.tester', 'testtest', use_current_url=True)
    assert browser.url.startswith(browser.PRINCIPAL_EDIT_URL_1)  # +/index.html


def test_principals__EditForm__7(
        address_book, UserFactory, localadmin, browser2):
    """Changing roles in `EditForm` changes the permissions of the user."""
    UserFactory(address_book, u'Urs', u'Unstable', u'uu@example.com',
                'u1u2u3u4', ['Visitor'])
    # As a visitor the user not able to create new persons:
    browser = browser2.formlogin(u'uu@example.com', 'u1u2u3u4')
    with pytest.raises(HTTPError) as err:
        browser.open(browser.PERSON_ADD_URL)
    assert 'HTTP Error 403: Forbidden' == str(err.value)
    # When the administrator changes the roles of a user ...
    localadmin.open(localadmin.PRINCIPAL_EDIT_URL)
    localadmin.getControl('roles').displayValue = ['Editor']
    localadmin.getControl('Apply').click()
    assert 'Data successfully updated.' == localadmin.message
    assert localadmin.PRINCIPALS_LIST_URL == localadmin.url
    # ... he immediately gets the new permissions:
    browser.open(browser.PERSON_ADD_URL)
    assert browser.getControl('first name')


def test_principals__EditForm__8(
        address_book, UserFactory, localadmin, browser2, browser3):
    """Removing all roles from a user leads to a Forbidden for this user."""
    UserFactory(address_book, u'Urs', u'Unstable', u'uu@example.com',
                'u1u2u3u4', ['Visitor'])
    browser = browser2.formlogin(u'uu@example.com', 'u1u2u3u4')
    # When the administrator removes all roles from a user ...
    browser.reload()
    url = browser.url
    localadmin.open(localadmin.PRINCIPAL_EDIT_URL)
    localadmin.getControl('roles').displayValue = []
    localadmin.getControl('Apply').click()
    assert 'Data successfully updated.' == localadmin.message
    assert localadmin.PRINCIPALS_LIST_URL == localadmin.url
    # ... he gets an exception that he is forbidden to see the page:
    with pytest.raises(HTTPError) as err:
        browser.open(url)
    assert 'HTTP Error 403: Forbidden' == str(err.value)
    # When he logs in again he gets an error message telling him that he has
    # not enough permissions to see anything:
    with pytest.raises(HTTPError) as err:
        browser3.formlogin(u'uu@example.com', 'u1u2u3u4')
    assert 'You have been logged-in successfully.' == browser3.message
    assert 'You are not authorized.' in browser.contents


def test_principals__DeleteUserForm__1(address_book, UserFactory, localadmin):
    """`DeleteUserForm` allows an administrator to delete a user."""
    UserFactory(address_book, u'Urs', u'Unstable', u'uu@example.com',
                'u1u2u3u4', ['Visitor'])
    browser = localadmin.open(localadmin.PRINCIPAL_EDIT_URL)
    browser.getControl('Delete user').click()
    assert browser.PRINCIPAL_DELETE_URL == browser.url
    browser.getControl('Yes').click()
    assert '"Unstable, Urs" deleted.' == browser.message
    assert browser.PRINCIPALS_LIST_URL == browser.url
    assert ['Admin, Arne'] == browser.etree.xpath('//td/a/text()')


def test_principals__DeleteUserForm__2(address_book, UserFactory, localadmin):
    """`DeleteUserForm` allows to cancel deletion of a user."""
    UserFactory(address_book, u'Urs', u'Unstable', u'uu@example.com',
                'u1u2u3u4', ['Visitor'])
    browser = localadmin.open(localadmin.PRINCIPAL_DELETE_URL)
    browser.getControl('No, cancel').click()
    assert 'Deletion canceled.' == browser.message
    assert browser.PRINCIPAL_EDIT_URL == browser.url


@pytest.mark.parametrize('username', ['visitor', 'editor'])
def test_principals__DeleteUserForm__3(
        address_book, UserFactory, browser, username):
    """Non-Admin users cannot access `DeleteUserForm`."""
    UserFactory(address_book, u'Urs', u'Unstable', u'uu@example.com',
                'u1u2u3u4', ['Visitor'])
    browser.login(username)
    with pytest.raises(HTTPError) as err:
        browser.open(browser.PRINCIPAL_DELETE_URL_1)
    assert 'HTTP Error 403: Forbidden' == str(err.value)



# -*- coding: utf-8 -*-
from .base import SessionStorageStep, get_fieldname_in_session
from icemac.addressbook.i18n import _
import collections
import decimal
import gocept.reference.field
import icemac.addressbook.fieldsource
import icemac.addressbook.sources
import sys
import z3c.form.field
import z3c.form.interfaces
import zc.sourcefactory.interfaces
import zope.component
import zope.interface
import zope.schema
import zope.schema.interfaces


class IWillNotGetStripped(zope.schema.interfaces.IText):
    """Suppress stripping of the field value in z3c.form."""


class NotStrippingTextConverter(z3c.form.converter.BaseDataConverter):
    """Since version 3.2.1 of z3c.form text values are stripped.

    This new feature is not wanted here. E. g. we want to be able to add text
    to a field prefixed with a newline.
    """

    zope.component.adapts(IWillNotGetStripped,
                          z3c.form.interfaces.IWidget)

    _strip_value = False


class IOperatorsSource(zc.sourcefactory.interfaces.IFactoredSource):
    """Marker interface for a source defining possible operators on a field."""


class BaseOperatorsSource(icemac.addressbook.sources.TitleMappingSource):
    """Base class for operator sources."""

    # grokcore.component does not work hier, don't know why
    zope.interface.implementsOnly(IOperatorsSource)


class TextOperatorsSource(BaseOperatorsSource):
    """Operators for Text and TextLine fields."""

    zope.component.adapts(zope.schema.interfaces.IText)

    _default_value = 'append'
    _missing_value = u''
    _mapping = collections.OrderedDict(
        (('prepend', _('prepend new value to existing one')),
         ('replace', _('replace existing value with new one')),
         ('append', _('append new value to existing one')),
         ('remove-all',
          _('remove all occurrences of new value in existing one')),
         ('remove-first',
          _('remove left-most occurrence of new value in existing one')),
         ('remove-last',
          _('remove right-most occurrence of new value in existing one')),
         ))


class ReplaceableOperatorsSource(BaseOperatorsSource):
    """Operators for Bool, Choice, Date, Datetime and URI fields."""

    _default_value = 'replace'
    _missing_value = None
    _mapping = collections.OrderedDict(
        (('replace', _('replace existing value with new one')),))


class KeywordOperatorsSource(BaseOperatorsSource):
    """Operators for keywords."""

    zope.component.adapts(gocept.reference.field.Set)

    _default_value = 'union'
    _missing_value = set()
    _mapping = collections.OrderedDict((
        ('union', _('append selected keywords to existing ones')),
        ('difference', _('remove selected keywords from existing ones')),
        ('intersection', _(
            'intersection (i. e. keywords which are in selected keywords and '
            'existing ones)')),
        ('symmetric_difference', _(
            'symmetric difference (i. e. keywords which are either in '
            'selected keywords or existing ones but not both)')),
    ))


class IntOperatorsSource(BaseOperatorsSource):
    """Operators for Int fields."""

    zope.component.adapts(zope.schema.interfaces.IInt)

    _default_value = 'add'
    _missing_value = 0
    _mapping = collections.OrderedDict(
        (('add', _('add new value to existing one')),
         ('sub', _('substract new value from existing one')),
         ('mul', _('multiply new value by existing one')),
         ('div', _('divide existing value by new one')),
         ('replace', _('replace existing value with new one')),
         ))


class DecimalOperatorsSource(IntOperatorsSource):
    """Operators for Decimal fields."""

    zope.component.adapts(zope.schema.interfaces.IDecimal)
    _missing_value = decimal.Decimal(0)


class Value(SessionStorageStep):
    """Step where the user enters the new value."""

    label = _(u'New value')
    fields = z3c.form.field.Fields()

    def _create_fields(self, session):
        entity, selected_field = icemac.addressbook.fieldsource.untokenize(
            session['field'])
        fields = []
        # We have to support user defined fields here:
        selected_field = zope.schema.interfaces.IField(selected_field)
        source = IOperatorsSource(selected_field)
        missing_value = source.factory._missing_value
        parameters = dict(
            title=_('new value'), required=False,
            description=_('This value should be set on each selected person.'),
            missing_value=missing_value)
        if isinstance(selected_field, zope.schema.Choice):
            # Choices need the source of selectable values:
            parameters['source'] = selected_field.source
        if isinstance(selected_field, gocept.reference.field.Set):
            # Keyword field needs its value_type:
            parameters['value_type'] = selected_field.value_type
        # If an IOrderable field has a missing_value which is not None,
        # min, max, and default have to be non-None values, too *sigh*:
        if zope.schema.interfaces.IInt.providedBy(selected_field):
            parameters.update(dict(min=-sys.maxint, max=sys.maxint, default=0))
        if zope.schema.interfaces.IDecimal.providedBy(selected_field):
            max_decimal = decimal.Decimal(sys.maxint)
            # The default value needs to be the same object as missing_value
            # as otherwise z3c.form displays the default value instead of
            # an empty field:
            parameters.update(dict(min=-max_decimal, max=max_decimal,
                                   default=missing_value))
        new_value_field = selected_field.__class__(**parameters)
        new_value_field.__name__ = get_fieldname_in_session(session['field'])
        if zope.schema.interfaces.IText.providedBy(new_value_field):
            zope.interface.alsoProvides(new_value_field, IWillNotGetStripped)
        fields.append(new_value_field)
        operation_field = zope.schema.Choice(
            title=_('operation'), source=source,
            description=_(
                'What should be done with the current value and the new one?'),
            default=source.factory._default_value, )
        operation_field.__name__ = 'operation'
        fields.append(operation_field)
        self.fields = z3c.form.field.Fields(*fields)

    def update(self):
        session = self.getContent()
        if 'field' in session:
            # Otherwise the user has selected this step before completing the
            # previous one.
            self._create_fields(session)
        super(Value, self).update()



from __future__ import print_function
from zope.app.publication.httpfactory import HTTPPublicationRequestFactory
import code
import fanstatic
import os
import sys
import zdaemon.zdctl
import zope.app.debug
import zope.app.wsgi
import zope.app.wsgi.interfaces
import zope.event

IGNORED = None


def zope_application_factory(
        global_conf=None, requestFactory=HTTPPublicationRequestFactory):
    """Create a Zope application."""
    if global_conf is None:
        db = None
    else:  # pragma: no cover (not testable)
        zope_conf = os.path.join(global_conf['here'], 'zope.conf')
        db = zope.app.wsgi.config(zope_conf)
    application = zope.app.wsgi.WSGIPublisherApplication(
        db, factory=requestFactory, handle_errors=True)
    return application


def application_factory(global_conf=None, zope_application=None):
    """Create an address book application with all WSGI middle wares.

    The concept and initial implementation came from
    `zope.app.wsgi.getWSGIApplication`.

    """
    if zope_application is None:  # pragma: no cover (not testable)
        zope_application = zope_application_factory(global_conf)
    application = fanstatic.make_fanstatic(
        zope_application, IGNORED, bottom=True, versioning=True,
        recompute_hashes=False)

    # Create the application, notify subscribers.
    zope.event.notify(
        zope.app.wsgi.interfaces.WSGIPublisherApplicationCreated(application))

    return application


def interactive_debug_prompt(zope_conf='zope.conf'):  # pragma: no cover
    db = zope.app.wsgi.config(zope_conf)
    debugger = zope.app.debug.Debugger.fromDatabase(db)
    # Invoke an interactive interpreter shell
    banner = ("Welcome to the interactive debug prompt.\n"
              "The 'root' variable contains the ZODB root folder.\n"
              "The 'app' variable contains the Debugger, 'app.publish(path)' "
              "simulates a request.")
    code.interact(banner=banner, local={
        'debugger': debugger, 'app': debugger, 'root': debugger.root()})


class ControllerCommands(zdaemon.zdctl.ZDCmd):  # pragma: no cover

    def do_debug(self, rest):
        interactive_debug_prompt()

    def help_debug(self):
        print("debug -- Initialize the application, providing a debugger")
        print("         object at an interactive Python prompt.")


def zdaemon_controller(zdaemon_conf='zdaemon.conf'):  # pragma: no cover
    args = ['-C', zdaemon_conf] + sys.argv[1:]
    zdaemon.zdctl.main(args, options=None, cmdclass=ControllerCommands)


def zdaemon_controller_debug_ajax():  # pragma: no cover (not testable)
    zdaemon_controller('zdaemon-debug-ajax.conf')


def zdaemon_controller_debug_pdb():  # pragma: no cover (not testable)
    zdaemon_controller('zdaemon-debug-pdb.conf')



from icemac.addressbook.interfaces import IKeywords, IKeyword
from icemac.addressbook.keyword import Keyword, KeywordContainer
import icemac.addressbook.interfaces
import icemac.addressbook.keyword
import zope.event
import zope.lifecycleevent


def test_keyword__Keyword__1():
    """`Keyword` fulfills the `IKeyword` interface."""
    zope.interface.verify.verifyObject(IKeyword, Keyword())


def test_keyword__changed__1(address_book, PersonFactory, KeywordFactory):
    """It updates the keyword index if a keyword title was changed."""
    friends = KeywordFactory(address_book, u'friends')
    person = PersonFactory(
        address_book, u'Kohn', keywords=[u'church', friends])

    catalog = zope.component.getUtility(zope.catalog.interfaces.ICatalog)
    results = catalog.searchResults(keywords={'any_of': ('friends', )})
    assert 1 == len(results)
    assert list(results)[0] is person

    # When the title of a keyword has changed the index gets updated at the
    # modified event:
    friends.title = u'Freunde'
    attributes = zope.lifecycleevent.Attributes(
        icemac.addressbook.interfaces.IKeyword, 'title')
    event = zope.lifecycleevent.ObjectModifiedEvent(friends, attributes)
    zope.event.notify(event)
    assert 2 == catalog.get('keywords').wordCount.value
    results = catalog.searchResults(keywords={'any_of': ('Freunde', )})
    assert 1 == len(results)
    assert list(results)[0] is person


def test_keyword__changed__2(address_book, PersonFactory, KeywordFactory):
    """It doesn't update the keyword index if keyword title wasn't changed."""
    friends = KeywordFactory(address_book, u'friends')
    person = PersonFactory(
        address_book, u'Kohn', keywords=[u'church', friends])

    catalog = zope.component.getUtility(zope.catalog.interfaces.ICatalog)
    results = catalog.searchResults(keywords={'any_of': ('friends', )})
    assert 1 == len(results)
    assert list(results)[0] is person

    # When the title of a keyword has not changed the index gets not updated at
    # the modified event:
    attributes = zope.lifecycleevent.Attributes(
        icemac.addressbook.interfaces.IKeyword, 'descr')
    event = zope.lifecycleevent.ObjectModifiedEvent(friends, attributes)
    zope.event.notify(event)
    results = catalog.searchResults(keywords={'any_of': ('friends', )})
    assert 1 == len(results)
    assert list(results)[0] is person


def test_keyword__KeywordContainer__1():
    """`KeywordContainer` fulfills the `IKeywords` interface."""
    zope.interface.verify.verifyObject(IKeywords, KeywordContainer())


def test_keyword__KeywordContainer__get_keywords__1():
    """It returns an empty generator if the container is empty."""
    assert [] == list(KeywordContainer().get_keywords())


def test_keyword__KeywordContainer__get_keywords__2():
    """It returns the keywords in the container."""
    kc = KeywordContainer()
    kc['1'] = Keyword(u'qwe')
    kc['2'] = Keyword(u'asd')
    kc['3'] = Keyword(u'dfg')
    kc['4'] = Keyword(u'bgr')
    keyword_titles = sorted(x.title for x in kc.get_keywords())
    assert ['asd', 'bgr', 'dfg', 'qwe'] == keyword_titles


def test_keyword__KeywordContainer__get_keyword_by_title__1():
    """It returns a keyword with a matching title."""
    kc = KeywordContainer()
    kc['1'] = Keyword(u'foo')
    kc['2'] = Keyword(u'bar')
    assert kc['1'] == kc.get_keyword_by_title(u'foo')
    assert kc['2'] == kc.get_keyword_by_title(u'bar')


def test_keyword__KeywordContainer__get_keyword_by_title__2():
    """It returns `None` if the container is empty."""
    assert None is KeywordContainer().get_keyword_by_title(u'asdf')


def test_keyword__KeywordContainer__get_keyword_by_title__3():
    """It returns `None` if no matching keyword can be found."""
    kc = KeywordContainer()
    kc['1'] = Keyword(u'asdf')
    assert None is kc.get_keyword_by_title(u'foo')
    assert None is kc.get_keyword_by_title(u'bar')


def test_keyword__KeywordContainer__get_keyword_by_title__4():
    """It returns the default value if there is no keyword for title."""
    kc = KeywordContainer()
    kc['1'] = Keyword(u'asdf')
    assert 'baz' == kc.get_keyword_by_title(u'foo', 'baz')



# -*- coding: utf-8 -*-
import fanstatic
import js.jquery
import js.jqueryui
import os.path
import zope.viewlet.viewlet


css_lib = fanstatic.Library('css', os.path.join('resources', 'css'))
js_lib = fanstatic.Library('js', os.path.join('resources', 'js'))


# CSS
reset_css = fanstatic.Resource(css_lib, 'reset.css')
base_css = fanstatic.Resource(css_lib, 'base.css', depends=[reset_css])
print_css = fanstatic.Resource(css_lib, 'print.css', depends=[base_css])
table_css = fanstatic.Resource(css_lib, 'table.css')
form_css = fanstatic.Resource(css_lib, 'form.css')
wizard_css = fanstatic.Resource(css_lib, 'wizard.css')
prefs_css = fanstatic.Resource(css_lib, 'prefs.css', depends=[form_css])
css = fanstatic.Group([
    base_css,
    print_css,
    form_css,
    prefs_css,
    table_css,
    wizard_css,
])

# not rendered on every page
no_max_content_css = fanstatic.Resource(
    css_lib, 'no_max_content.css', depends=[form_css])
background_css = fanstatic.Resource(
    css_lib, 'background.css', depends=[base_css])


# JavaScript
masterdata_fields = fanstatic.Resource(
    js_lib, 'masterdata_fields.js', bottom=True,
    depends=[js.jqueryui.ui_sortable])
table = fanstatic.Resource(
    js_lib, 'table.js', bottom=True, depends=[js.jquery.jquery])
prefs = fanstatic.Resource(
    js_lib, 'prefs.js', bottom=True, depends=[js.jqueryui.effects_fade])
form = fanstatic.Resource(
    js_lib, 'form.js', bottom=True, depends=[js.jquery.jquery])


js = fanstatic.Group([
    form,
    masterdata_fields,
    prefs,
    table,
])


class DefaultResources(zope.viewlet.viewlet.ViewletBase):
    """Resources which are needed for each page."""

    def update(self):
        css.need()
        js.need()

    def render(self):
        return u''


class AddressBookResources(zope.viewlet.viewlet.ViewletBase):
    """Resources which are address book specific."""

    def update(self):
        background_css.need()

    def render(self):
        return u''



import icemac.addressbook.generations.utils
import icemac.addressbook.interfaces
import zope.annotation.interfaces
import zope.component
import zope.interface
import zope.location


@icemac.addressbook.generations.utils.evolve_addressbooks
def evolve(ab):
    """Update order storage to make it metadata aware."""
    os = zope.component.getUtility(
        icemac.addressbook.interfaces.IOrderStorage)
    for ns in os.namespaces():
        storage = os.byNamespace(ns)
        zope.interface.alsoProvides(
            storage, zope.annotation.interfaces.IAttributeAnnotatable)
        zope.location.locate(storage, os._storage, ns)



# -*- coding: latin-1 -*-
import icemac.addressbook.generations.utils


generation = 3


def evolve(context):
    """Install the authentication utility."""
    icemac.addressbook.generations.utils.update_address_book_infrastructure(
        context)



# -*- coding: latin-1 -*-
import icemac.addressbook.generations.utils

generation = 1


def evolve(context):
    """Install the necessary components for address books."""
    icemac.addressbook.generations.utils.update_address_book_infrastructure(
        context)



from z3c.form.i18n import MessageFactory as _
import icemac.addressbook.browser.base


class Redirect(icemac.addressbook.browser.base.BaseView):
    """Redirect back to person list"""

    def __call__(self):
        self.send_flash(_('Data successfully updated.'))
        self.request.response.redirect(
            self.url(self.context, 'person-list.html'))



# -*- coding: utf-8 -*-
from icemac.addressbook.entities import Field
from icemac.addressbook.interfaces import IField
from zope.interface.exceptions import Invalid
from zope.schema import getValidationErrors
import pytest


@pytest.fixture('function')
def field():
    """Get a preconfigured field object."""
    field = Field()
    field.title = u'my field'
    field.order = 1.5
    return field


def test_interfaces__IField__1(field):
    """It does not raise `Invalid` if the field type is not 'choice'."""
    field.type = u'Text'
    errors = getValidationErrors(IField, field)
    assert [] == errors


def test_interfaces__IField__2(field):
    """Using the `choice` type requires values to be added."""
    field.type = u'Choice'
    errors = getValidationErrors(IField, field)
    assert 1 == len(errors)
    assert errors[0][0] is None
    assert isinstance(errors[0][1], Invalid)
    assert ((u'type "choice" requires at least one field value.',) ==
            errors[0][1].args)


def test_interfaces__IField__3(field):
    """It does not raise `Invalid` if there are values for a choice field."""
    field.type = u'Choice'
    field.values = [u'sure', u'not really']
    errors = getValidationErrors(IField, field)
    assert [] == errors



# -*- coding: utf-8 -*-
from .roles import has_editor_role, has_visitor_role
from icemac.addressbook.i18n import _
import gocept.reference
import icemac.addressbook.interfaces
import icemac.addressbook.principals.interfaces
import zope.annotation.interfaces
import zope.container.interfaces
import zope.interface
import zope.lifecycleevent
import zope.pluggableauth.plugins.principalfolder
import zope.securitypolicy.interfaces


PASSWORD_MANAGER_NAME = 'SSHA'
EDITOR_VISITOR_PERMS = ('icemac.addressbook.EditPrincipalPassword',
                        'icemac.addressbook.ViewPrincipal',)
ONLY_EDITOR_PERMS = ('icemac.addressbook.EditPrincipal', )


class Principal(zope.pluggableauth.plugins.principalfolder.InternalPrincipal):
    """Principal where the password manager cannot be specified."""

    zope.interface.implements(
        icemac.addressbook.principals.interfaces.IPrincipal,
        icemac.addressbook.principals.interfaces.IPasswordFields,
        icemac.addressbook.principals.interfaces.IRoles,
        zope.annotation.interfaces.IAttributeAnnotatable)

    _person = gocept.reference.Reference('_person', ensure_integrity=True)
    _login = None
    _password = None
    description = None
    _roles = ()
    last_login = None

    def __init__(self):
        self._passwordManagerName = PASSWORD_MANAGER_NAME

    @property
    def title(self):
        # needed by principal folder
        return icemac.addressbook.interfaces.ITitle(self.person)

    @property
    def person(self):
        return self._person

    @person.setter
    def person(self, person):
        if self._person is not None:
            # it's not possible to change the person
            raise ValueError(
                'It is not possible to change the person on a principal.')
        self._person = person
        self.login = person.default_email_address.email

    @property
    def password(self):
        return self._password

    @password.setter
    def password(self, password):
        if password is not None:
            zope.pluggableauth.plugins.principalfolder.InternalPrincipal.\
                setPassword(self, password)

    @property
    def password_repetition(self):
        pass  # password_repetition is not stored

    @password_repetition.setter
    def password_repetition(self, password):
        pass  # password_repetition is not stored

    @property
    def roles(self):
        return self._roles

    @roles.setter
    def roles(self, roles):
        old_roles = self._roles
        self._roles = roles
        addressbook = icemac.addressbook.principals.interfaces.IRoot(self)
        # Update role map on addressbook
        role_manager = (
            zope.securitypolicy.interfaces.IPrincipalRoleManager(
                addressbook))
        for role in old_roles:
            role_manager.unsetRoleForPrincipal(role, self.__name__)
        for role in roles:
            role_manager.assignRoleToPrincipal(role, self.__name__)

        # Update permission map on user
        permission_manager = (
            zope.securitypolicy.interfaces.IPrincipalPermissionManager(
                self))

        for permission in EDITOR_VISITOR_PERMS + ONLY_EDITOR_PERMS:
            permission_manager.unsetPermissionForPrincipal(
                permission, self.__name__)
        editor_role = has_editor_role(roles)
        visitor_role = has_visitor_role(roles)
        if editor_role or visitor_role:
            for permission in EDITOR_VISITOR_PERMS:
                permission_manager.grantPermissionToPrincipal(
                    permission, self.__name__)
        if editor_role:
            for permission in ONLY_EDITOR_PERMS:
                permission_manager.grantPermissionToPrincipal(
                    permission, self.__name__)


@zope.component.adapter(
    icemac.addressbook.principals.interfaces.IPrincipal,
    zope.lifecycleevent.IObjectCreatedEvent)
def created(principal, event):
    """Create initial infrastructure."""
    # Set a default value here for the reference as z3c.form accesses the
    # attributes before a value is assigned and gets an AttributeError
    # otherwise:
    assert not hasattr(principal, '_person')
    principal._person = None


@zope.component.adapter(
    icemac.addressbook.principals.interfaces.IPrincipal,
    zope.container.interfaces.IObjectAddedEvent)
def added(principal, event):
    # roles need principal.__name__ so we set them again when the
    # principal has been added
    principal.roles = principal._roles


@zope.component.adapter(icemac.addressbook.principals.interfaces.IPrincipal)
@zope.interface.implementer(icemac.addressbook.interfaces.ITitle)
def title(principal):
    if getattr(principal, 'person', None) is not None:
        return icemac.addressbook.interfaces.ITitle(principal.person)
    # safety belt:
    return _(u'<no person>')



# -*- coding: utf-8 -*-
import ConfigParser
import os.path
import shutil
import subprocess
import sys

sys.path[0:0] = ['src']

import icemac.addressbook.install  # noqa


USER_INI = 'install.user.ini'


def run_process(text, *args):
    """Run a script in a subprocess."""
    print '%s ...' % text
    res = subprocess.call(args)
    if res:
        sys.exit(res)


def bool_get(config, key):
    """Read value from config."""
    value = config.get('migration', key)
    return value == 'yes'


def copy_dir(src_base, dest_base, *path_parts):
    """Copy directory from src_base + path_parts to dest_base + path_parts."""
    path = os.path.join(*path_parts)
    src_dir = os.path.join(src_base, path)
    dest_dir = os.path.join(dest_base, path)
    if os.path.exists(dest_dir):
        shutil.rmtree(dest_dir)
    shutil.copytree(src_dir, dest_dir)


def migrate():
    """Migrate an old address book instance."""
    # Read the ini file the configurator just created to get the
    # migration options.
    config = ConfigParser.SafeConfigParser()
    config.read(USER_INI)

    if not bool_get(config, 'do_migration'):
        # no migration wanted
        return
    old_instance = config.get('migration', 'old_instance')
    if not (old_instance and
            os.path.exists(os.path.join(old_instance, USER_INI))):
        print 'ERROR: You did not provide a path to the old instance.'
        print '       So I can not migrate the old content.'
        sys.exit(-1)
    cwd = os.getcwd()
    try:
        os.chdir(old_instance)
        demon_path = os.path.join('bin', 'addressbook')
        if bool_get(config, 'stop_server'):
            run_process('Stopping old instance', demon_path, 'stop')
        run_process('Packing the old instance keeping 1 day of old data',
                    os.path.join(cwd, 'bin', 'offlinepack'), '--days=1',
                    os.path.join('var', 'Data.fs'))
        run_process('Creating backup of old instance',
                    os.path.join('bin', 'snapshotbackup'))
        print 'Copying data backups to new instance ...'
        copy_dir(old_instance, cwd, 'var', 'snapshotbackups')
        print 'Copying blob backups to new instance ...'
        copy_dir(old_instance, cwd, 'var', 'blobstoragesnapshots')
    finally:
        os.chdir(cwd)

    run_process('Restoring backup into new instance',
                os.path.join('bin', 'snapshotrestore'), '--no-prompt')
    if bool_get(config, 'start_server'):
        run_process('Starting new instance', demon_path, 'start')


if __name__ == '__main__':
    python = sys.executable
    if icemac.addressbook.install.not_matched_prerequisites():
        print icemac.addressbook.install.not_matched_prerequisites()
        sys.exit(-1)
    conf_args = []
    if len(sys.argv) > 1:
        conf_args.append(os.path.join(sys.argv[1], USER_INI))
    icemac.addressbook.install.Configurator(*conf_args)()

    run_process('running %s bootstrap.py' % python, python, 'bootstrap.py',
                '--setuptools-version=20.3.1', '--buildout-version=2.5.0')
    run_process('running bin/buildout', 'bin/buildout')
    migrate()

    print 'Installation complete.'



# -*- coding: utf-8 -*-
from icemac.addressbook.i18n import MessageFactory as _
import icemac.addressbook.principals.sources
import zope.interface
import zope.schema


class IPasswordFields(zope.interface.Interface):
    """Required password fields to enter and check password."""

    password = zope.schema.Password(
        title=_(u'password'), min_length=8,
        description=_(u'The password for the user.'))

    password_repetition = zope.schema.Password(
        title=_(u'password repetition'), min_length=8,
        description=_(u'Please repeat the password.'))

    @zope.interface.invariant
    def password_eq_repetition(obj):
        if obj.password != obj.password_repetition:
            raise zope.interface.Invalid(
                _(u'Entry in password field was not equal to entry in '
                  u'password repetition field.'))


class IPrincipal(zope.interface.Interface):
    """Derived from ...

    ... zope.pluggableauth.plugins.principalfolder.IInternalPrincipal.
    """

    last_login = zope.interface.Attribute('Datetime of last login or `None`.')

    person = zope.schema.Choice(
        title=_(u'person'), readonly=True,
        source=icemac.addressbook.principals.sources.persons)

    login = zope.schema.TextLine(
        title=_(u'login name'),
        description=_(u'The Login/Username of the user. '
                      u'This value can change.'))

    description = zope.schema.Text(
        title=_('notes'),
        description=_('Provides notes for the user.'),
        required=False, missing_value='', default=u'')


class IRoles(zope.interface.Interface):
    """Roles"""

    roles = zope.schema.Tuple(
        title=_(u'roles'), required=False,
        value_type=zope.schema.Choice(
            source=icemac.addressbook.principals.sources.role_source))


class IRoot(zope.interface.Interface):
    """Root object on which global roles are stored.

    You have to provide an adapter from the root object to this interface.
    """


class IVisitorRole(zope.interface.Interface):
    """Role which is set on a user who is only allowed to read."""


class IEditorRole(zope.interface.Interface):
    """Role which is set on a user who is only allowed to read and write."""



import icemac.addressbook.generations.utils


def evolve(context):
    """Install update default preferences to get time zone in."""
    icemac.addressbook.generations.utils.update_address_book_infrastructure(
        context)



# -*- coding: utf-8 -*-
# Copyright (c) 2013-2014 Michael Howitz
from ..interfaces import IIconProviderInfo
import icemac.addressbook
import pkg_resources
import zope.contentprovider.provider


class About(object):
    """View class for the about screen."""

    def version(self):
        return pkg_resources.get_distribution('icemac.addressbook').version

    def icons(self):
        return zope.component.subscribers([None], IIconProviderInfo)


class CopyrightContentProvider(
        zope.contentprovider.provider.ContentProviderBase):
    """Content provider for the copyright string."""

    def render(self):
        return icemac.addressbook.copyright


class IconProviderInfo(object):
    """Infos about copyrights of icons used in the address book."""
    zope.interface.implements(IIconProviderInfo)

    name = NotImplemented
    url = NotImplemented

    def __init__(self, *ignored):
        pass


class DialogIcons(IconProviderInfo):
    name = u'VistaICO.com'
    url = (u'http://www.vistaico.com')


class AddressBookIcon(IconProviderInfo):
    name = u'Martin Šnajdr'
    url = (u'http://psd.tutsplus.com/tutorials/designing-tutorials/'
           u'create-a-custom-mac-osx-style-ring-binder-address-book-icon')



# -*- coding: utf-8 -*-
import decimal
import gocept.reference.collection
import grokcore.component
import zope.interface


class IOperator(zope.interface.Interface):
    """Update operator."""

    def __init__(operand1):
        """Store operand one."""

    def __call__(operand2):
        """Operate `operand2` on `operand1`."""


class Operator(grokcore.component.Adapter):
    """Base class for operators."""

    grokcore.component.implements(IOperator)
    grokcore.component.context(object)

    def __init__(self, operand1):
        self.operand1 = operand1

    def __call__(self, operand2):
        raise NotImplementedError()


class TextAppend(Operator):
    """Append to text."""

    grokcore.component.context(unicode)
    grokcore.component.name('append')

    def __call__(self, operand2):
        return self.operand1 + operand2


class Replace(Operator):
    """Replace."""

    grokcore.component.context(zope.interface.Interface)
    grokcore.component.name('replace')

    def __call__(self, operand2):
        return operand2


class NoneAppend(Replace):
    """Append to None."""

    grokcore.component.name('append')


class TextPrepend(Operator):
    """Prepend to text."""

    grokcore.component.context(unicode)
    grokcore.component.name('prepend')

    def __call__(self, operand2):
        return operand2 + self.operand1


class NonePrepend(Replace):
    """Prepend to None."""

    grokcore.component.name('prepend')


class RemoveAll(Operator):
    """Remove all occurrences of operand2 from operand1."""

    grokcore.component.context(unicode)
    grokcore.component.name('remove-all')

    def __call__(self, operand2):
        return self.operand1.replace(operand2, '')


class NoneRemoveAll(Operator):
    """Remove all for None."""

    grokcore.component.context(zope.interface.Interface)
    grokcore.component.name('remove-all')

    def __call__(self, operand2):
        return self.operand1


class RemoveFirst(Operator):
    """Remove left-most occurrence of operand2 from operand1."""

    grokcore.component.context(unicode)
    grokcore.component.name('remove-first')

    def __call__(self, operand2):
        return self.operand1.replace(operand2, '', 1)


class NoneRemoveFirst(NoneRemoveAll):
    """Remove first for None."""

    grokcore.component.name('remove-first')


class RemoveLast(Operator):
    """Remove right-most occurrence of operand2 from operand1."""

    grokcore.component.context(unicode)
    grokcore.component.name('remove-last')

    def __call__(self, operand2):
        # [::-1] reverses the string
        return self.operand1[::-1].replace(operand2[::-1], '', 1)[::-1]


class NoneRemoveLast(NoneRemoveAll):
    """Remove last for None."""

    grokcore.component.name('remove-last')


class KeywordOperator(Operator):
    """Base class for keyword operators."""

    grokcore.component.context(gocept.reference.collection.InstrumentedSet)

    def __init__(self, operand1):
        super(KeywordOperator, self).__init__(operand1)
        # convert InstrumentedSet to a set so normal set operations are
        # useable on it
        self.operand1 = set(x for x in self.operand1)


class KeywordUnion(KeywordOperator):
    """Union keywords."""

    grokcore.component.name('union')

    def __call__(self, operand2):
        return self.operand1.union(operand2)


class KeywordDifference(KeywordOperator):
    """Difference keywords."""

    grokcore.component.name('difference')

    def __call__(self, operand2):
        return self.operand1.difference(operand2)


class KeywordIntersection(KeywordOperator):
    """Intersect keywords."""

    grokcore.component.name('intersection')

    def __call__(self, operand2):
        return self.operand1.intersection(operand2)


class KeywordSymmetricDifference(KeywordOperator):
    """Compute symmetric difference of keywords."""

    grokcore.component.name('symmetric_difference')

    def __call__(self, operand2):
        return self.operand1.symmetric_difference(operand2)


class IntAdd(Operator):
    """Add operand2 to operand1 for ints."""

    grokcore.component.context(int)
    grokcore.component.name('add')

    def __call__(self, operand2):
        return self.operand1 + operand2


class DecimalAdd(IntAdd):
    """Add operand2 to operand1 for decimals."""

    grokcore.component.context(decimal.Decimal)


class NoneAdd(Operator):
    """Add for None."""

    grokcore.component.context(zope.interface.Interface)
    grokcore.component.name('add')

    def __call__(self, operand2):
        return operand2


class IntSub(Operator):
    """Substract operand2 from operand1 for ints."""

    grokcore.component.context(int)
    grokcore.component.name('sub')

    def __call__(self, operand2):
        return self.operand1 - operand2


class DecimalSub(IntSub):
    """Substract operand2 from operand1 for decimals."""

    grokcore.component.context(decimal.Decimal)


class NoneSub(Operator):
    """Sub for None."""

    grokcore.component.context(zope.interface.Interface)
    grokcore.component.name('sub')

    def __call__(self, operand2):
        return -operand2


class IntMul(Operator):
    """Multiply operand2 by operand1 for ints."""

    grokcore.component.context(int)
    grokcore.component.name('mul')

    def __call__(self, operand2):
        return self.operand1 * operand2


class DecimalMul(IntMul):
    """Multiply operand2 by operand1 for decimals."""

    grokcore.component.context(decimal.Decimal)


class NoneMul(Operator):
    """Mul for None."""

    grokcore.component.context(zope.interface.Interface)
    grokcore.component.name('mul')

    def __call__(self, operand2):
        return 0


class IntDiv(Operator):
    """Divide operand1 by operand2 for ints."""

    grokcore.component.context(int)
    grokcore.component.name('div')

    def __call__(self, operand2):
        return self.operand1 / operand2


class DecimalDiv(IntDiv):
    """Divide operand1 by operand2 for decimals."""

    grokcore.component.context(decimal.Decimal)


class NoneDiv(Operator):
    """Div for None."""

    grokcore.component.context(zope.interface.Interface)
    grokcore.component.name('div')

    def __call__(self, operand2):
        return 0



from ..base import can_access_uri_part
from icemac.addressbook.i18n import _
import icemac.addressbook.browser.base
import icemac.addressbook.interfaces
import zope.dottedname.resolve
import zope.interface


class StartpageData(object):
    """Data of a startpage."""

    zope.interface.implements(icemac.addressbook.interfaces.IStartpageData)

    def __init__(self, iface_name, view, title):
        assert view is not None  # required by `can_access_uri_part`, see below
        self.iface_name = iface_name
        self.view = view
        self.title = title

    def __call__(self, *args):
        return self


welcome = StartpageData(
    'icemac.addressbook.interfaces.IAddressBook', 'welcome.html',
    _('Welcome page'))
person_list = StartpageData(
    'icemac.addressbook.interfaces.IAddressBook', 'person-list.html',
    _('Person list'))
search = StartpageData(
    'icemac.addressbook.interfaces.IAddressBook', 'search.html',
    _('Search'))


class Dispatch(icemac.addressbook.browser.base.BaseView):
    """Dispatch to the selected start page."""

    def _get_context_and_view_name(self, src):
        iface_name, view_name = src
        iface = zope.dottedname.resolve.resolve(iface_name)
        context = iface(self.context)
        return context, view_name

    def __call__(self):
        context, view_name = self._get_context_and_view_name(
            self.context.startpage)
        if not can_access_uri_part(context, self.request, view_name):
            context, view_name = self._get_context_and_view_name(
                icemac.addressbook.interfaces.DEFAULT_STARTPAGE_DATA)
        target_url = self.url(context, view_name)
        self.request.response.redirect(target_url)
        return ''



from icemac.addressbook.i18n import _
import icemac.addressbook.interfaces
import zc.sourcefactory.basic
import zope.component


def tokenize(entity, field_name):
    """Convert an entity and a field_name into a unique string token."""
    return "%s###%s" % (entity.name, field_name)


def untokenize(token):
    """Convert a token containing entity and field name back to the objects."""
    entity_name, field_name = token.split('###')
    entity = icemac.addressbook.interfaces.IEntity(entity_name)
    field = entity.getRawField(field_name)
    return entity, field


class Source(zc.sourcefactory.basic.BasicSourceFactory):
    """Fields of a person."""

    def getValues(self):
        # Only the main entities (the ones with default values on the
        # person) are in this source as only their presence is guaranteed.
        entities = zope.component.getUtility(
            icemac.addressbook.interfaces.IEntities)
        for entity in entities.getMainEntities():
            # All fields, even the user defined ones, can show up here.
            for field_name, field in entity.getRawFields():
                # We return a representation here which can be stored in
                # ZODB even though it is not so easy to compute a title from
                # it. But it would be more difficult to adapt consumers of
                # this source (e. g. zope.preference) to store a string
                # instead of non-persistent field objects.
                yield tokenize(entity, field_name)

    def getTitle(self, value):
        entity, field = untokenize(value)
        # The titles might be message ids so allow to translate them.
        return _(u"${prefix} -- ${title}", mapping=dict(
            prefix=entity.title, title=field.title))


source = Source()



# -*- coding: utf-8 -*-
import icemac.addressbook.browser.table
import icemac.addressbook.fieldsource
import icemac.addressbook.interfaces
import icemac.truncatetext
import z3c.form.term
import z3c.table.column
import zope.component
import zope.i18n
import zope.preference.interfaces
import zope.schema.interfaces


class LinkColumn(z3c.table.column.LinkColumn,
                 icemac.addressbook.browser.table.BaseColumn):
    """LinkColumn which fetches link content from item."""

    def getSortKey(self, item):
        # Sort by the value displayed to the user, but lowercase it,
        # as the user does not expect lower case names below the
        # 'normal' upper case ones.
        return self.getLinkContent(item).lower()

    def getLinkContent(self, item):
        # Get the content from item or return the default value when empty.
        return self.getValue(item)

    def renderCell(self, item):
        if self.getLinkContent(item):
            return super(LinkColumn, self).renderCell(item)
        return self.defaultValue


class URLColumn(LinkColumn):
    """LinkColumn where the link URL is the same as the link content."""

    getLinkURL = LinkColumn.getLinkContent


class TruncatedContentColumn(icemac.addressbook.browser.table.BaseColumn):
    """Column which truncates its content."""

    length = 20  # number of characters to display
    ellipsis = u'…'  # ellipsis sign

    def renderCell(self, item):
        return icemac.truncatetext.truncate(
            self.getRawValue(item), self.length, self.ellipsis)


class BoolColumn(icemac.addressbook.browser.table.BaseColumn):
    """AdaptedGetAttrColumn for displaying bool values."""

    def renderCell(self, item):
        value = self.getRawValue(item)
        # We use the labels of z3c.form here so the displayed values
        # are the same as in the edit form.
        if value is True:
            label = z3c.form.term.BoolTerms.trueLabel
        elif value is False:
            label = z3c.form.term.BoolTerms.falseLabel
        else:
            return self.defaultValue
        return zope.i18n.translate(label, context=self.request)


class TranslatedTiteledColumn(icemac.addressbook.browser.table.BaseColumn):
    """Column which returns the translated ITitle of the value."""

    def renderCell(self, obj):
        """Get the title of the value."""
        value = self.getRawValue(obj)
        translated = self.defaultValue
        if value:
            title = icemac.addressbook.interfaces.ITitle(value)
            translated = zope.i18n.translate(title, context=self.request)
        return translated


class EMailColumn(icemac.addressbook.browser.table.BaseColumn,
                  z3c.table.column.EMailColumn):
    """Column which renders the cell contents as mailto-link."""

    renderCell = z3c.table.column.EMailColumn.renderCell


USER_DEFINED_FIELD_TYPE__TO__COLUMN = {
    u'Bool': BoolColumn,
    u'URI': URLColumn,
    u'Text': TruncatedContentColumn,
    u'Date': icemac.addressbook.browser.table.DateColumn,
    u'Datetime': icemac.addressbook.browser.table.DateTimeColumn,
}


FIELD_NAME__TO__COLUMN = {
    'first_name': LinkColumn,
    'last_name': LinkColumn,
    'keywords': icemac.addressbook.browser.table.KeywordsColumn,
    'country': TranslatedTiteledColumn,
    'email': EMailColumn,
    'url': URLColumn,
}


def getColumnClass(entity, field):
    """Get a column class to display the requested field of an entity."""
    column_class = FIELD_NAME__TO__COLUMN.get(field.__name__, None)
    if column_class is not None:
        return column_class
    if (zope.schema.interfaces.IText.providedBy(field) and
            not zope.schema.interfaces.ITextLine.providedBy(field)):
        # The content of text areas (but not text lines, which extend from
        # text area) should get truncated.
        return TruncatedContentColumn
    if zope.schema.interfaces.IDate.providedBy(field):
        # Date fields need a special column, as None values are not
        # compareable to date values:
        return icemac.addressbook.browser.table.DateColumn
    if icemac.addressbook.interfaces.IField.providedBy(field):
        return USER_DEFINED_FIELD_TYPE__TO__COLUMN.get(
            field.type, icemac.addressbook.browser.table.BaseColumn)
    # Field is e. g. a TextLine
    return icemac.addressbook.browser.table.BaseColumn


def createFieldColumn(table, entity, field, weight):
    """Create a column for the specified `entity` and `field`."""
    additional_column_args = dict()
    if (field.__name__ == 'url' or
        (icemac.addressbook.interfaces.IField.providedBy(field) and
         field.type == 'URI')):
        additional_column_args['linkTarget'] = '_blank'

    return z3c.table.column.addColumn(
        table, getColumnClass(entity, field), field.__name__,
        header=field.title, weight=weight,
        entity=entity, field=field, **additional_column_args)


class BasePersonList(object):
    """List persons in address book without batching.

    Either extend `icemac.addressbook.browser.table.PageletTable`
    or `icemac.addressbook.browser.table.Table`.
    """

    def __init__(self, context, request, prefs=None):
        super(BasePersonList, self).__init__(context, request)
        self.sortOrder = self.prefs.personLists.sort_direction

    @property
    def prefs(self):
        """User defined preferences for person list."""
        return zope.component.getUtility(
            zope.preference.interfaces.IPreferenceGroup, name="ab")

    def update(self):
        self._columns = self._set_up_columns()
        super(BasePersonList, self).update()

    def _set_up_columns(self):
        """Create the columns of the table."""
        zcml_columns = super(BasePersonList, self).setUpColumns()
        zcml_columns_before = len(
            [x for x in zcml_columns if z3c.table.table.getWeight(x) < 0])
        python_columns = self._set_up_user_selected_columns_and_sort_on(
            zcml_columns_before)
        return zcml_columns + python_columns

    def _set_up_user_selected_columns_and_sort_on(self, columns_before):
        """Create the columns selected by the user in the preferences.

        Additionally computes the sort order.
        """
        order_by = self.prefs.personLists.order_by
        columns = []
        index = 0  # current column index
        # Entity and field of the column which should be used for order-by:
        try:
            order_by_entity, order_by_field = (
                icemac.addressbook.fieldsource.untokenize(order_by))
        except KeyError:
            # Field has been deleted, so we can't use it for sorting:
            order_by_entity, order_by_field = None, None

        # Create all columns and compute the order-by column:
        for column_name in self.prefs.personLists.columns:
            try:
                entity, field = (
                    icemac.addressbook.fieldsource.untokenize(column_name))
            except KeyError:
                # Column no longer exists
                continue
            columns.append(createFieldColumn(self, entity, field, index))
            if entity == order_by_entity and field == order_by_field:
                # Found the entity and field for order by, so set the
                # expected name on the sortOn variable. The last parameter
                # depends on the position of the column in the table when
                # there are columns before the columns defined by this
                # method the index gets higher:
                self.sortOn = '%s-%s-%s' % (
                    self.prefix, field.__name__, index + columns_before)
            # current column exists, so the index can be adjusted
            index += 1
        return columns

    def setUpColumns(self):
        """Return the computed columns and via ZCML registered ones."""
        return self._columns

    values = NotImplemented



from .interfaces import IEditor
from datetime import datetime
from zope.dublincore.interfaces import IZopeDublinCore
import icemac.addressbook.metadata.storage
import pytest
import zope.interface.verify


@pytest.yield_fixture(scope='function')
def person(address_book, browser):
    """Create a person using a browser so `creator` is set."""
    browser.login('editor')
    browser.open(browser.PERSON_ADD_URL)
    browser.getControl('last name').value = 'Tester'
    browser.getControl('Add').click()
    assert '"Tester" added.' == browser.message
    # Set address_book as site again as using a browser resets it.
    with zope.component.hooks.site(address_book):
        yield address_book['Person']


def test_storage__EditorMetadataStorage__1():
    """`EditorMetadataStorage` conforms to `IEditor`."""
    assert zope.interface.verify.verifyObject(
        IEditor,
        icemac.addressbook.metadata.storage.EditorMetadataStorage())


def test_subscriber__1(address_book, PersonFactory):
    """Zope sets the creation date on creation."""
    person = PersonFactory(address_book, u'Tester')
    assert isinstance(IZopeDublinCore(person).created, datetime)


def test_subscriber__CreatorAnnotator__1(person):
    """`CreatorAnnotator` sets the creator on adding a person with browser."""
    assert u'global editor' == IEditor(person).creator


def test_subscriber__LastModifierAnnotator__1(person):
    """`LastModifierAnnotator` sets creator as last modifier at creation."""
    assert u'global editor' == IEditor(person).modifier


def test_subscriber__LastModifierAnnotator__2(
        address_book, person, UserFactory, browser2):
    """`LastModifierAnnotator` sets last modifier at modification."""
    # Users inside the address book are valid modifiers, too.
    UserFactory(address_book, u'Hans', u'Tester', u'hans@test.de',
                u'12345678', ['Editor'])
    browser = browser2.formlogin('hans@test.de', '12345678')
    browser.open(browser.PERSON_EDIT_URL)
    browser.getControl('notes').value = 'I was here.'
    browser.getControl('Apply').click()
    assert 'Data successfully updated.' == browser.message
    assert u'global editor' == IEditor(person).creator
    assert u'Tester, Hans' == IEditor(person).modifier



import icemac.addressbook.principals.interfaces
import zope.component


def editor_role(ignored):
    return 'icemac.addressbook.global.Editor'


def visitor_role(ignored):
    return 'icemac.addressbook.global.Visitor'


def _has_role(roles, role_interface):
    """Tell whether an a role specified by `role_interface` is in `roles`."""
    possible_roles = set(zope.component.subscribers((None,), role_interface))
    return not possible_roles.isdisjoint(set(roles))


def has_editor_role(roles):
    """Tell whether an editor role is in `roles`."""
    return _has_role(
        roles, icemac.addressbook.principals.interfaces.IEditorRole)


def has_visitor_role(roles):
    """Tell whether an editor role is in `roles`."""
    return _has_role(
        roles, icemac.addressbook.principals.interfaces.IVisitorRole)



import zope.generations.utility
import icemac.addressbook.interfaces
import icemac.addressbook.addressbook
import logging


logger = logging.getLogger('evolve')


def evolve_addressbooks(func):
    """Decorator which evolves address books."""
    def decorated(context):
        root = zope.generations.utility.getRootFolder(context)
        addressbooks = zope.generations.utility.findObjectsProviding(
            root, icemac.addressbook.interfaces.IAddressBook)
        for addressbook in addressbooks:
            logger.info('evolving %r' % addressbook)
            old_site = zope.component.hooks.getSite()
            try:
                zope.component.hooks.setSite(addressbook)
                func(addressbook)
            finally:
                zope.component.hooks.setSite(old_site)
    return decorated


@evolve_addressbooks
def update_address_book_infrastructure(addressbook):
    """Update the address book infrastructure (e. g. install new utilities)."""
    icemac.addressbook.addressbook.create_address_book_infrastructure(
        addressbook)



import grokcore.component as grok
import icemac.addressbook.browser.menus.interfaces
import itertools
import z3c.menu.ready2go.checker
import z3c.menu.ready2go.manager
import zope.viewlet.manager
import zope.viewlet.viewlet


MainMenu = zope.viewlet.manager.ViewletManager(
    'main-menu', icemac.addressbook.browser.menus.interfaces.IMainMenu,
    bases=(z3c.menu.ready2go.manager.MenuManager,))


def getWeight((name, viewlet)):
    view_name = viewlet.viewName
    assert view_name.startswith('@@')
    view_name = view_name[2:]
    view = zope.component.getMultiAdapter(
        (viewlet.context, viewlet.request), name=view_name)
    interface = getattr(view, 'interface', None)
    if interface is None:
        interface = getattr(view, 'interface_for_menu', None)
    assert interface is not None, (
        'View %r does neither have a non-null value on `interface` nor '
        'on `interface_for_menu` attribute. But this is needed to compute '
        'the AddMenu contents.' % view)
    entity = icemac.addressbook.interfaces.IEntity(interface)
    order = zope.component.getUtility(
        icemac.addressbook.interfaces.IEntityOrder)
    try:
        return order.get(entity)
    except (KeyError, zope.component.ComponentLookupError):
        # The entity is not known in the order or we are outside an address
        # book.
        return 0


class OrdersWeightMenuManager(z3c.menu.ready2go.manager.MenuManager):
    """Menu manager which uses address_book.orders as weight."""

    def sort(self, viewlets):
        return sorted(viewlets, key=getWeight)


class EmptyViewlet(zope.viewlet.viewlet.ViewletBase):
    """Helper class for AlwaysRenderTemplateManager."""


EMPTY_VIEWLET = EmptyViewlet(None, None, None, None)


class AlwaysRenderTemplateManager(OrdersWeightMenuManager):
    """Manager rendering template even there are no viewlets."""

    def update(self):
        super(AlwaysRenderTemplateManager, self).update()
        self.have_viewlets = bool(self.viewlets)
        if not self.have_viewlets:
            # Trick our `render` method to render manager's template which
            # should check `self.have_viewlets` to not render the empty
            # viewlet.
            self.viewlets.append(EMPTY_VIEWLET)


AddMenu = zope.viewlet.manager.ViewletManager(
    'add-menu', icemac.addressbook.browser.menus.interfaces.IAddMenu,
    bases=(AlwaysRenderTemplateManager,))


class SelectMenuItemOn(object):
    """Subscription adapter factory

    Returns interfaces or names of views for which a specific menu item in the
    main menu should be selected.

    Expects view names without `@@`.
    Example usage::

      my_search_views = SelectMenuItemOn('foosearch.html', IMyObject)

      <subscriber
         for="*"
         provides=".interfaces.ISearchMenuItemOn"
         factory=".my_search_views" />
    """

    def __init__(self, *items):
        self.items = items

    def __call__(self, *args):
        return self.items


class SubscriberSelectedChecker(
        z3c.menu.ready2go.checker.ViewNameSelectedChecker,
        grok.MultiAdapter):
    """Selected checker using a subscription adapter to check view names.

    In subclass as a ``grok.adapts`` statement like this::

        grok.adapts(zope.interface.Interface,
                    icemac.addressbook.browser.interfaces.IAddressBookLayer,
                    zope.interface.Interface,
                    icemac.addressbook.browser.menus.menu.MainMenu,
                    <concrete menu item class>)
    """

    subscriber_interface = NotImplemented
    grok.baseclass()

    @property
    def selected(self):
        if super(SubscriberSelectedChecker, self).selected:
            return True
        view_name = self.view.__name__
        context = self.context
        items = zope.component.subscribers((self,), self.subscriber_interface)
        for item in itertools.chain(*items):
            if isinstance(item, basestring):
                if view_name == item:
                    return True
            # If item is no view name it has to be an interface:
            elif item.providedBy(context):
                return True
        return False



import icemac.addressbook.browser.menus.menu
import icemac.addressbook.browser.search.result.handler.base
import icemac.addressbook.interfaces


class MailTo(icemac.addressbook.browser.search.result.handler.base.Base):
    """Create a mailto link for all persons in selection."""

    @property
    def unique_mail_addresses(self):
        addresses = set()
        for person in self.persons:
            email = icemac.addressbook.interfaces.IEMailAddress(person, None)
            if email is None:
                continue
            if not email.email:
                continue
            addresses.add(email.email)
        return sorted(addresses)

    def mailto_link(self):
        return 'mailto:?bcc=%s' % ','.join(self.unique_mail_addresses)


mailto = icemac.addressbook.browser.menus.menu.SelectMenuItemOn(
    'mailto.html')



from icemac.addressbook.i18n import _
import gocept.reference.interfaces
import icemac.addressbook.browser.base
import icemac.addressbook.browser.interfaces
import icemac.addressbook.browser.menus.menu
import icemac.addressbook.browser.metadata
import icemac.addressbook.browser.table
import icemac.addressbook.interfaces
import z3c.form.button
import z3c.table.column
import zope.interface


@zope.interface.implementer(
    icemac.addressbook.browser.interfaces.IAddressBookBackground)
class AddForm(icemac.addressbook.browser.base.BaseAddForm):

    label = _(u'Add new keyword')
    interface = icemac.addressbook.interfaces.IKeyword
    class_ = icemac.addressbook.keyword.Keyword
    next_url = 'parent'


def can_delete_keyword(form):
    """Button condition telling if the displayed keyword is deleteable."""
    return (
        icemac.addressbook.browser.base.can_access('delete.html')(form) and
        not gocept.reference.interfaces.IReferenceTarget(
            form.context).is_referenced())


@zope.interface.implementer(
    icemac.addressbook.browser.interfaces.IAddressBookBackground)
class EditForm(icemac.addressbook.browser.base.GroupEditForm):

    groups = (icemac.addressbook.browser.metadata.MetadataGroup,)
    label = _(u'Edit keyword')
    interface = icemac.addressbook.interfaces.IKeyword
    next_url = 'parent'
    z3c.form.form.extends(icemac.addressbook.browser.base.GroupEditForm,
                          ignoreFields=True)

    def applyChanges(self, data):
        # GroupForm has its own applyChanges but we need the one from
        # _AbstractEditForm here as inside the goups no changes are made but
        # there is a subscriber which raises an error which is handled by
        # _AbstractEditForm.
        return icemac.addressbook.browser.base._AbstractEditForm.applyChanges(
            self, data)

    @z3c.form.button.buttonAndHandler(
        _(u'Delete'), name='delete', condition=can_delete_keyword)
    def handleDelete(self, action):
        self.redirect_to_next_url('object', 'delete.html')


@zope.interface.implementer(
    icemac.addressbook.browser.interfaces.IAddressBookBackground)
class DeleteForm(icemac.addressbook.browser.base.BaseDeleteForm):
    label = _(u'Do you really want to delete this keyword?')
    interface = icemac.addressbook.interfaces.IKeyword
    field_names = ('title', )


@zope.interface.implementer(
    icemac.addressbook.browser.interfaces.IAddressBookBackground)
class Table(icemac.addressbook.browser.table.Table):
    """List keywords in address book."""

    no_rows_message = _(u'No keywords defined yet.')

    def setUpColumns(self):
        """Return the previously computed columns."""
        return [z3c.table.column.addColumn(
            self, icemac.addressbook.browser.table.TitleLinkColumn, 'keyword',
            header=_(u'keyword')),
        ]

    @property
    def values(self):
        """The values are stored on the context."""
        return self.context.get_keywords()


keyword_views = icemac.addressbook.browser.menus.menu.SelectMenuItemOn(
    icemac.addressbook.interfaces.IKeyword,
    icemac.addressbook.interfaces.IKeywords, 'addKeyword.html')



# Starting point of data serialization for full text index, not complete, yet


def FieldSerializer(field, obj):
    """Serialize the value of a field to a unicode value."""
    return unicode(field.get(obj))


def FieldNoSerializer(field, obj):
    """Serialize the value of a field to the empty string."""
    return u''



import icemac.addressbook.generations.utils
import zope.component
import zope.principalannotation.interfaces


@icemac.addressbook.generations.utils.evolve_addressbooks
def evolve(addressbook):
    """Copy global principal annotations to local ones."""
    global_ann = zope.component.getUtility(
        zope.principalannotation.interfaces.IPrincipalAnnotationUtility,
        context=addressbook.__parent__)
    local_ann = zope.component.getUtility(
        zope.principalannotation.interfaces.IPrincipalAnnotationUtility,
        context=addressbook)
    principals = zope.component.getUtility(
        zope.pluggableauth.interfaces.IAuthenticatorPlugin,
        name=u'icemac.addressbook.principals')

    for pid, ann in global_ann.annotations.items():
        if pid not in principals:
            # principal does not exist in this address book
            continue
        local_ann.annotations[pid] = ann



"""Base classes and functions for wizards."""

import z3c.form.field
import z3c.wizard.step
import z3c.wizard.wizard


class Step(z3c.wizard.step.Step):
    """Customized wizard step which (easier to use)."""

    @property
    def fields(self):
        return z3c.form.field.Fields(self.interface)


class Wizard(z3c.wizard.wizard.Wizard):
    """Base (customizable) wizard."""



import BTrees.OOBTree
import icemac.addressbook.interfaces
import icemac.addressbook.utils
import persistent
import persistent.list
import zope.annotation.interfaces
import zope.container.contained
import zope.interface
import zope.lifecycleevent
import zope.location


@zope.interface.implementer(zope.annotation.interfaces.IAttributeAnnotatable)
class OrderStorageList(persistent.list.PersistentList):
    """List storing orders of a single namespace.

    This list is metadata aware.
    """


class OrderStorage(persistent.Persistent,
                   zope.container.contained.Contained):
    """Storage of orders of objects."""

    zope.interface.implements(icemac.addressbook.interfaces.IOrderStorage)

    def __init__(self, *args, **kw):
        self._storage = icemac.addressbook.utils.create_obj(
            BTrees.OOBTree.OOBTree)

    # IOrderStorageRead
    def namespaces(self):
        """Get an iterable of the known namespaces."""
        return self._storage.keys()

    def byNamespace(self, namespace):
        """Get the objects in an order namespace as a list."""
        return self._storage[namespace]

    def get(self, obj, namespace):
        """Get the index of the object in the list."""
        try:
            return self.byNamespace(namespace).index(obj)
        except ValueError:
            raise KeyError(obj)

    def isFirst(self, obj, namespace):
        """Tell whether `obj` is the first object in the list."""
        return self.get(obj, namespace) == 0

    def isLast(self, obj, namespace):
        """Tell whether `obj` is the last object in the list."""
        return ((self.get(obj, namespace) + 1) ==
                len(self.byNamespace(namespace)))

    # IOrderStorageWrite

    def add(self, obj, namespace):
        """Add an object to the order for a namespace."""
        if namespace not in self._storage:
            self._create_namespace(namespace)
        storage = self.byNamespace(namespace)
        if obj not in storage:
            storage.append(obj)
            self._modified(storage)

    def remove(self, obj, namespace):
        """Remove the object from the order of a namespace."""
        storage = self.byNamespace(namespace)
        storage.remove(obj)
        self._modified(storage)

    def truncate(self, namespace):
        """Remove all objects from the order of a namespace."""
        if namespace in self._storage:
            storage = self.byNamespace(namespace)
            # Remove all contents from storage:
            del storage[:]
            self._modified(storage)

    def up(self, obj, namespace, delta=1):
        """Move the object one position up in the list."""
        storage = self.byNamespace(namespace)
        for _ in range(delta):
            index = storage.index(obj)
            if index == 0:
                raise ValueError(
                    "Moving %r by %s positions up would move it beyond the "
                    "beginning of the list." % (obj, delta))
            storage[index - 1:index + 1] = reversed(
                storage[index - 1:index + 1])
        self._modified(storage)

    def down(self, obj, namespace, delta=1):
        """Move the object one position down in the list."""
        storage = self.byNamespace(namespace)
        for i in xrange(delta):
            index = storage.index(obj)
            if index == len(storage) - 1:
                raise ValueError(
                    "Moving %r by %s positions down would move it beyond the "
                    "end of the list." % (obj, delta))
            storage[index:index + 2] = reversed(storage[index:index + 2])
        self._modified(storage)

    # private

    def _create_namespace(self, namespace):
        """Create a new order namespace."""
        namespace_storage = icemac.addressbook.utils.create_obj(
            OrderStorageList)
        self._storage[namespace] = namespace_storage
        zope.location.locate(namespace_storage, self._storage, namespace)

    def _modified(self, obj):
        zope.lifecycleevent.modified(obj)



from __future__ import absolute_import
import icalendar
import pkg_resources


def Calendar(**kw):
    """Factory to create an icalendar.Calendar with address book defaults."""
    calendar = icalendar.Calendar()
    calendar.add(
        'prodid', '-//icemac.addressbook//%s//EN' %
        pkg_resources.get_distribution('icemac.addressbook').version)
    calendar.add('version', '2.0')
    calendar.add('calscale', 'GREGORIAN')
    for key, value in kw.items():
        calendar.add(key, value)
    return calendar


def set_download_request_headers(request, filename_base):
    """Set the needed headers on the request to download an iCalendar file.

    filename_base ... filename without extension.

    """
    request.response.setHeader('Content-Type', 'text/calendar')
    request.response.setHeader('Content-Disposition',
                               'attachment; filename=%s.ics' % filename_base)



from icemac.addressbook.i18n import _
import gocept.reference.interfaces
import icemac.addressbook.browser.base
import icemac.addressbook.browser.interfaces
import icemac.addressbook.browser.metadata
import icemac.addressbook.interfaces
import icemac.addressbook.person
import icemac.addressbook.utils
import transaction
import z3c.form.button
import z3c.form.datamanager
import z3c.form.form
import z3c.form.group
import zc.sourcefactory.contextual
import zope.interface


class AddGroup(icemac.addressbook.browser.base.PrefixGroup):
    """PrefixGroup for AddForm."""

    def __init__(self, context, request, parent, interface, label, prefix):
        super(AddGroup, self).__init__(context, request, parent)
        self.interface = interface
        self.label = label
        self.prefix = prefix


class PersonEditGroup(AddGroup):
    """PrefixGroup for addresses IPerson in EditForm."""

    def __init__(self, context, request, parent, interface, label, prefix,
                 index, key):
        super(PersonEditGroup, self).__init__(
            context, request, parent, interface, label, prefix)
        self.prefix = "%s_%s" % (prefix, index)
        self.key = key

    def update(self):
        self.groups = [
            icemac.addressbook.browser.metadata.MetadataGroup(
                self.getContent(), self.request, self)]
        super(PersonEditGroup, self).update()


class AddressEditGroup(PersonEditGroup):
    """PrefixGroup for addresses in EditForm."""

    def getContent(self):
        return self.context[self.key]


class FileEditGroup(AddressEditGroup):
    """EditGroup for icemac.addressbook.file.interfaces.IFile objects."""


class DefaultSelectGroup(icemac.addressbook.browser.base.PrefixGroup):
    """Group to select the default addresses."""

    interface = icemac.addressbook.interfaces.IPersonDefaults
    label = icemac.addressbook.person.person_defaults_entity.title
    prefix = 'defaults'


@zope.interface.implementer(
    icemac.addressbook.browser.interfaces.IAddressBookBackground)
class PersonAddForm(z3c.form.group.GroupForm,
                    icemac.addressbook.browser.base.BaseAddForm):
    """Add a person."""

    label = _(u'Add new person')
    next_url = 'parent'
    next_view = 'person-list.html'
    interface_for_menu = icemac.addressbook.interfaces.IPerson

    def __init__(self, *args, **kw):
        super(PersonAddForm, self).__init__(*args, **kw)
        context = self.context
        request = self.request
        entities = zope.component.getUtility(
            icemac.addressbook.interfaces.IEntities)
        groups = [AddGroup(context, request, self, entity.interface,
                           entity.title, entity.name)
                  for entity in entities.getMainEntities()]
        self.groups = tuple(groups)

    def createAndAdd(self, data):
        # Create person first
        person_entity = icemac.addressbook.interfaces.IEntity(
            icemac.addressbook.interfaces.IPerson)
        person_entity_name = person_entity.name
        person_group = [group
                        for group in self.groups
                        if group.prefix == person_entity_name][0]
        person = icemac.addressbook.browser.base.create(
            person_group, person_entity.getClass(), data)
        self._name = icemac.addressbook.utils.add(self.context, person)
        for group in self.groups:
            if group.prefix == person_entity_name:
                # already created above
                continue
            entity = icemac.addressbook.interfaces.IEntity(group.prefix)
            obj = icemac.addressbook.browser.base.create(
                group, entity.getClass(), data)
            icemac.addressbook.utils.add(person, obj)
            # handling of default addresses: the first address is
            # saved as default:
            setattr(person, entity.tagged_values['default_attrib'], obj)
        self.obj = person
        return person


def person_deletable(form):
    """Button display constraint checking whether a person is deleteable."""
    person = form.context
    ref_target = gocept.reference.interfaces.IReferenceTarget(person)
    return not ref_target.is_referenced(recursive=False)


class PersonEditForm(icemac.addressbook.browser.base.GroupEditForm):

    label = _(u'Edit person data')
    next_url = 'parent'
    next_view = 'person-list.html'
    interface_for_menu = icemac.addressbook.interfaces.IPerson

    def __init__(self, *args, **kw):
        super(PersonEditForm, self).__init__(*args, **kw)
        entity_util = zope.component.getUtility(
            icemac.addressbook.interfaces.IEntities)
        entities = entity_util.getMainEntities(sorted=False)
        file_entity = icemac.addressbook.interfaces.IEntity(
            icemac.addressbook.file.interfaces.IFile)
        defaults_entity = icemac.addressbook.interfaces.IEntity(
            icemac.addressbook.interfaces.IPersonDefaults)
        entities.extend([file_entity, defaults_entity])

        groups = []
        for entity in icemac.addressbook.entities.sorted_entities(entities):
            index = 0

            # Special handling for IPerson
            if entity.interface == icemac.addressbook.interfaces.IPerson:
                group = PersonEditGroup(
                    self.context, self.request, self, entity.interface,
                    entity.title, entity.name, index,
                    self.context.__name__)
                groups.append(group)
                continue

            # Special handling for IPersonDefaults
            IPersonDefaults = icemac.addressbook.interfaces.IPersonDefaults
            if entity.interface == IPersonDefaults:
                groups.append(
                    DefaultSelectGroup(self.context, self.request, self))
                continue

            is_ifile = (
                entity.interface == icemac.addressbook.file.interfaces.IFile)
            if is_ifile:
                # IFile needs a special group and has a fix title
                group_class = FileEditGroup
                obj_title = entity.title
            else:
                # for the other entities the title depends on whether being
                # main address or not
                default_attrib = entity.tagged_values.get('default_attrib')
                default_obj = getattr(self.context, default_attrib)
                group_class = AddressEditGroup

            for obj in icemac.addressbook.utils.iter_by_interface(
                    self.context, entity.interface):
                if not is_ifile:
                    if obj == default_obj:
                        obj_title = _(u'main ${address}')
                    else:
                        obj_title = _(u'other ${address}')
                    obj_title = _(obj_title, mapping={'address': entity.title})
                group = group_class(
                    self.context, self.request, self, entity.interface,
                    obj_title, entity.name, index, obj.__name__)
                groups.append(group)
                index += 1

        self.groups = tuple(groups)

    @z3c.form.button.buttonAndHandler(_('Apply'), name='apply')
    def handleApply(self, action):
        super(PersonEditForm, self).handleApply(self, action)
        # update file's mime type after new file was uploaded
        for group in self.groups:
            if not isinstance(group, FileEditGroup):
                continue
            prefix = z3c.form.util.expandPrefix(group.prefix)
            icemac.addressbook.browser.file.file.update_blob(
                group.widgets[prefix + 'data'], group.getContent())

    @z3c.form.button.buttonAndHandler(
        _(u'Clone person'), name='clone_person',
        condition=icemac.addressbook.browser.base.can_access('clone.html'))
    def handleClonePerson(self, action):
        self.redirect_to_next_url('object', 'clone.html')

    @z3c.form.button.buttonAndHandler(
        _(u'Delete whole person'), name='delete_person',
        condition=icemac.addressbook.browser.base.all_(
            icemac.addressbook.browser.base.can_access('delete_person.html'),
            person_deletable))
    def handleDeletePerson(self, action):
        self.redirect_to_next_url('object', 'delete_person.html')

    @z3c.form.button.buttonAndHandler(
        _(u'Delete single entry'), name='delete_entry',
        condition=icemac.addressbook.browser.base.can_access(
            'delete_entry.html'))
    def handleDeleteAddress(self, action):
        self.redirect_to_next_url('object', 'delete_entry.html')

    @z3c.form.button.buttonAndHandler(
        _(u'Export'), name='export',
        condition=icemac.addressbook.browser.base.can_access(
            'export.html'))
    def handleExport(self, action):
        self.redirect_to_next_url('object', 'export.html')

    def applyChanges(self, data):
        """Special variant which sends ModifiedEvent for each modified object.

        (not to be used universally!)
        """
        content = self.getContent()
        changed = {}
        assert self.fields.items() == []  # Each field is in a group.
        for group in self.groups:
            groupChanged = group.applyChanges(data)
            if groupChanged:
                changed[group.getContent()] = groupChanged
        for content, changes in changed.items():
            descriptions = [
                zope.lifecycleevent.Attributes(interface, *names)
                for interface, names in changes.items()]
            zope.event.notify(
                zope.lifecycleevent.ObjectModifiedEvent(
                    content, *descriptions))
        return changed


class KeywordDataManager(z3c.form.datamanager.AttributeField):
    """Datamanager which converts the internal InstrumentedSet into a set.

    This way it is comparable with selected values.
    """

    def get(self):
        return set(x for x in super(KeywordDataManager, self).get())


class DeletePersonForm(icemac.addressbook.browser.base.BaseDeleteForm):
    label = _(u'Do you really want to delete this person?')
    interface = icemac.addressbook.interfaces.IPerson
    field_names = ('first_name', 'last_name')
    next_view_after_delete = 'person-list.html'

    def _do_delete(self):
        try:
            super(DeletePersonForm, self)._do_delete()
        except gocept.reference.interfaces.IntegrityError:
            transaction.abort()
            message = _(
                'Failed to delete person: This person is referenced. '
                'To delete this person, remove the reference before.')
            zope.component.getUtility(
                z3c.flashmessage.interfaces.IMessageSource).send(message)


class ClonePersonForm(icemac.addressbook.browser.base.BaseCloneForm):
    label = _(u'Do you really want to clone this person?')
    interface = icemac.addressbook.interfaces.IPerson
    field_names = ('first_name', 'last_name')


class PersonEntriesSource(
        zc.sourcefactory.contextual.BasicContextualSourceFactory):
    """Source to select entries (addresses, files, ...) from a person."""

    def getValues(self, context):
        entities = zope.component.getUtility(
            icemac.addressbook.interfaces.IEntities).getEntities()
        for entity in entities:
            assert entity.interface
            source = icemac.addressbook.interfaces.ContextByInterfaceSource(
                entity.interface).factory.getValues(context)
            for value in source:
                yield value

    def getTitle(self, context, value):
        entity = icemac.addressbook.interfaces.IEntity(value)
        title = _('${prefix} -- ${title}',
                  mapping=dict(
                      prefix=entity.title,
                      title=icemac.addressbook.interfaces.ITitle(value)))
        return title


class IPersonEntries(zope.interface.Interface):
    """Content entries of a person."""

    entry = zope.schema.Choice(
        title=_(u'Entries'), source=PersonEntriesSource())


class DeleteSingleEntryForm(icemac.addressbook.browser.base.BaseEditForm):
    """Form to choose entry for deletion."""

    label = _(u'Please choose an entry for deletion:')
    interface = IPersonEntries
    ignoreContext = True
    next_url = 'object'

    @z3c.form.button.buttonAndHandler(_(u'Delete entry'), name='delete_entry')
    def handleDeleteEntry(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        selected_entry = data['entry']

        self.request.response.redirect(self.url(selected_entry, 'delete.html'))
